/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 * This code is part of EXP4.
 *
 * This module builds and evaluates the krylov space by Arnoldi recursions
 *
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <float.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 20 */
#define LOCAL_DEBUG_OFFSET 6
#include "debug.h"

#include "pfuns.h"
#include "work.h"
#include "krylov.h"
#include "scnorm.h"
#include "krystep.h"

#include "arnoldi.h"

static finteger inc0 = 0;
static finteger inc1 = 1;

static Complex wone  = ComplexConst(1.0);
static Complex wzero = ComplexConst(0.0);

static void PREC(Exp4SizeArnoldiWork) (krylov *Kry,size_t *Shared,
      size_t *Resident);

static WorkSpace PREC(Exp4InitArnoldi) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,krylov *Kry);

static void PREC(Exp4IncreaseDimensionArnoldi) (WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *X,finteger ldX,Complex *Z,finteger ldZ,
      finteger fac_t,Complex *Start,finteger UpTo,krylov *Kry);

static void PREC(Exp4KrylovCombineLinearArnoldi) (Complex *x,
      finteger ldx,Complex *Start,krylov *Kry,
      Complex *zz,finteger ldz,finteger fac_t,int dummy,int dummy2);


/* creates a struct krylov which holds data for an Arnoldi process */
void PREC(Exp4FillHandleWithArnoldi) (struct Krylov *Kry,finteger MaxKrylovDim,
      finteger neq,int HaveOwnV)
{
  /* set the functions for use with this data */
  Kry->KrylovData[KrylovActive(Kry)].KrylovSizeWorkFun =
    & PREC(Exp4SizeArnoldiWork);
  Kry->KrylovData[KrylovActive(Kry)].InitKryFun =
    & PREC(Exp4InitArnoldi);
  Kry->KrylovData[KrylovActive(Kry)].KrylovGetHGuessFun = NULL;
  Kry->KrylovData[KrylovActive(Kry)].KrylovIncreaseDimensionFun =
    & PREC(Exp4IncreaseDimensionArnoldi);
  Kry->KrylovData[KrylovActive(Kry)].KrylovCombineLinearFun =
    & PREC(Exp4KrylovCombineLinearArnoldi);

  KrylovMaxSize(Kry)    = MaxKrylovDim;
  KrylovHBandWith(Kry)  = MaxKrylovDim;

  if (HaveOwnV)
    ArnoldiVData(Kry) = Exp4GetWorkPointer(sizeof(Complex)*
	  KrylovLdBig(Kry)*KrylovMaxSize(Kry));

  Exp4Message1((DEBUG_FILE,"Created Arnoldi Handle: Maximum Dimension = %d,"
	" %s\n",KrylovMaxSize(Kry),
	(KrylovCanRebuild(Kry))?"Own V workspace":"Dynamic V workspace"));
}

static void PREC(Exp4SizeArnoldiWork) (krylov *Kry,size_t *Shared,
      size_t *Resident)
{
  (*Shared) = GetWorkSpaceSize(sizeof(Complex)*KrylovNeq(Kry));

  if (KrylovCanRebuild(Kry))
    (*Resident) = (size_t)0;
  else
    (*Resident) = GetWorkSpaceSize(sizeof(Complex)*
	  KrylovNeq(Kry)*KrylovMaxSize(Kry))
      +           GetWorkSpaceSize(sizeof(Complex)*
	    KrylovLdSmall(Kry)*KrylovMaxSize(Kry));
}

static WorkSpace PREC(Exp4InitArnoldi) (WorkSpace Work,Complex *Start,
      Exp4KrylovInitType What,krylov *Kry)
{
  switch (What) {
  case Build:
    {
      if (!KrylovCanRebuild(Kry))
      {
	size_t SizeV,SizeH;
	/* adjust the maximum krylov dimension for Arnoldi according
	   to the avaliable memory */
	int NVec = (GetRemainingWorkSpaceSize(Work)-KrylovReserveSize(Kry))/
	  (sizeof(Complex)*KrylovNeq(Kry));
	KrylovMaxSize(Kry) = min(NVec,KrylovLdSmall(Kry));

	Exp4Message2((DEBUG_FILE,"Set Maximum Krylov Dimension of dynamic "
	      "Krylov Process to %d\n",KrylovMaxSize(Kry)));

	/* distribute workspace for H and V */
	SizeV = sizeof(Complex)*KrylovLdBig(Kry)*KrylovMaxSize(Kry);
	SizeH = sizeof(Complex)*KrylovLdSmall(Kry)*KrylovMaxSize(Kry);

	DistributeWorkspace(ArnoldiVData(Kry),SizeV,Work);
	DistributeWorkspace(KrylovH(Kry),SizeH,Work);
      }

      /* normalize the first vector in the Krylov base */
      pcopy (&KrylovNeq(Kry),Start,&inc1,ArnoldiV(Kry,0),&inc1);
      prscl (&KrylovNeq(Kry),&KrylovNormv(Kry),ArnoldiV(Kry,0),&inc1);
      break;
    }
  case ReBuild:
    {
      if (!KrylovCanRebuild(Kry))
	ErrorReturn("Can not rebuild unknown krylov space!",
	      EXP4_ERR_TRY_TO_REBUILD_UNKNOWN_KRYLOV_SPACE);
      break;
    }
  }

  return Work;
}

static void PREC(Exp4IncreaseDimensionArnoldi)(WorkSpace Work,
      Real *ResiualNorm,int *AcceptKonvergence,
      Complex *X,finteger ldX,Complex *Z,finteger ldZ,
      finteger fac_t,Complex *Start,finteger UpTo,krylov *Kry)
{
  Complex *Tmp,*Vkp1;

  DistributeWorkspace(Tmp,sizeof(Complex)*KrylovNeq(Kry),Work);
  Vkp1 = ArnoldiV(Kry,KrylovSize(Kry));

  while(KrylovSize(Kry) < UpTo)
  {
#if (defined(SINGLE_PRECISION) || defined(SINGLE_COMPLEX))
    Real normVkp1 = FLT_MAX;
#else
    Real normVkp1 = DBL_MAX;
#endif
    finteger i;
    finteger k = KrylovSize(Kry);
    finteger km1 = k-1;
    finteger kp1 = k+1;
    
    /* initialize the KrylovSize(Kry)-th line with zeros */
    if (1 < k)
      pcopy(&km1,&wzero,&inc0,&KrylovHent(Kry,k,0),&KrylovLdSmall(Kry));

    /* normalize the Base vectors, the Start vector is already normalized */
    if (1 <= k)
      prscl (&KrylovNeq(Kry),&Re(KrylovHent(Kry,k,km1)),ArnoldiV(Kry,k),&inc1);

    Vkp1 = (kp1<KrylovMaxSize(Kry))?ArnoldiV(Kry,kp1):Tmp;

    /* Vkp1 = A*V(:,k); */
    PREC(Exp4Mv) (KrylovNeq(Kry),Vkp1,ArnoldiV(Kry,k),1);
    
    for (i=0;i<=k;i++)
    {
      Complex whelp;

      /* h_nkp1=V(:,j)'*tmp */
      pdot (&whelp,&KrylovNeq(Kry),ArnoldiV(Kry,i),&inc1,Vkp1,&inc1);

      KrylovHent(Kry,i,k) = whelp;

      /* Vkp1 = Vkp1-V(:,j)*h_nkp1 */
      neg(whelp,whelp);
      paxpy (&KrylovNeq(Kry),&whelp,ArnoldiV(Kry,i),&inc1,Vkp1,&inc1);
    }

    /* normalize the Base vectors, the Start vector is already normalized */
    if (kp1 < KrylovMaxSize(Kry)) {
      normVkp1 = pnrm2 (&KrylovNeq(Kry),ArnoldiV(Kry,kp1),&inc1);
    
      /* Generate (KrylovSize(Kry)) -st line of H */
      ComplexSetReal(KrylovHent(Kry,kp1,k),normVkp1);
    }

    KrylovSize(Kry)++;

    /* if an invariant subspace is detected FIXME */
    if (rabs(normVkp1) < 1e-10) break;
  }  /* while(KrylovSize(Kry) < UpTp) */  

  (*ResiualNorm) = PREC(Exp4ScaledNorm) (Vkp1);
  (*AcceptKonvergence) = 1;
}

static void PREC(Exp4KrylovCombineLinearArnoldi) (Complex *X,
      finteger ldX,Complex *Start,krylov *Kry,
      Complex *Z,finteger ldZ,finteger fac_t,int dummy,int dummy2)
{
/*    int i; */
  
/*    for (i=0;i<fac_t;i++) */
/*      pgemv ("No transpose",&KrylovNeq(Kry),&KrylovSize(Kry), */
/*   	  &wone,ArnoldiV(Kry,0),&KrylovLdBig(Kry),Z+i*ldZ,&inc1, */
/*   	  &wzero,X+i*ldX,&inc1); */

  /* x = V(:,1:k)*z(1:k,:) */
  if (fac_t == 1) {
    pgemv ("No transpose",&KrylovNeq(Kry),&KrylovSize(Kry),
	  &wone,ArnoldiV(Kry,0),&KrylovLdBig(Kry),Z,&inc1,
	  &wzero,X,&inc1);
  } else {
    pgemm ("No transpose","No transpose",&KrylovNeq(Kry),&fac_t,
	  &(KrylovSize(Kry)),&wone,ArnoldiV(Kry,0),&KrylovLdBig(Kry),
	  Z,&ldZ,&wzero,X,&ldX);
  }
}
