/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
   The stepsize control module for the core integrator determined by the
   local error estimation of the code
 */

/* a little workaround for a (possible ?) bug in gcc-2.7.2.3 or
 * libc.5.4.33 headers from linux (my machine at home ... ) */
#ifdef  __GNUC__
#define __NO_MATH_INLINES
#endif

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include "ftypes.h"
#include "prec.h"
#include "exp4.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 15 */
#define LOCAL_DEBUG_OFFSET 4
#include "debug.h"

#include "errstep.h"

/* the smallest accepted stepsize change */
#define FACMIN                0.2
/* the largest accepted stepsize change */
#define FACMAX                5.0
/* the security factor */
#define FAC                   0.9
/* the lower bound for accepted errors, to avoid confusing Gustafson */
#define MIN_ERR_ACC           1e-2


/* sets the the error and the stepsize of the last accepted step */
void PREC(ErrstepSetAccepted) (Exp4ErrorStepSizeControlHandle *Handle,
      Real err,Real h)
{
  Handle->h_acc   = h;
  Handle->err_acc = max(MIN_ERR_ACC,err);

  Exp4Message1((DEBUG_FILE,"Accepted h and err: h = %10.4e, err = "
	"%10.4e\n",Handle->h_acc,Handle->err_acc));
}

/* returns the stepsize change ratio for the Gustafson approach */
static Real PREC(Gustafson) (Real ThisStepsize,Real ThisError,
      Real LastError,Real LastStepsize,Real ROrder)
{
  return ThisStepsize/LastStepsize
    * pow(LastError/(ThisError*ThisError), ROrder);
}

/* returns the stepsize change ratio for the Classical approach */
static Real PREC(Classical) (const Real ThisError, const Real ROrder)
{
  return pow(ThisError, - ROrder);
}

/*
 * multiplies the ratio with the security factor and
 * clips the stepsize ratio to the interval [FACMIN, FACMAX]
 */
static Real PREC(ClipFactor) (Real f)
{
  /* apply the secutity factor */
  f *= FAC;
  /* clip against maximum */
  f = min(f, FACMAX);
  /* clip against minimum */
  f = max(f, FACMIN);
  
  return f;
}

/* calculates the stepsize for the next step */
Real PREC(ErrstepGetNewH) (const Exp4ErrorStepSizeControlHandle *Handle,
      const Real ThisError,const Real h,const int order)
{
  Real h_err;

  if ((ThisError <= 1.0) && (0.0 < Handle->h_acc)) {
    /* Stepsize control by the minimum of Gustafson and Classical,
     * taken from dopri. */
    Real f_gus   = PREC(Gustafson) (h, ThisError,
                      Handle->err_acc, Handle->h_acc, 1.0/order );
    Real f_class = PREC(Classical) (ThisError, 1.0/order );

    h_err = h*PREC(ClipFactor) (min(f_gus, f_class));

    Exp4Message2((DEBUG_FILE,"Take Gustafson for stepsize control\n"));
  } else {
    /* Stepsize control by Classical if we are in the first step,
     * or this setp will be rejected. */

    h_err = h*PREC(ClipFactor) (PREC(Classical) (ThisError,1.0/order));
  }

  Exp4Message1((DEBUG_FILE,"Stepsize control: new stepsize: %e\n",h_err));

  return h_err;
}
