/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 EXP4:   Solve large systems of stiff ordinary differential equations
         in autonomous form.

 CODE:   This code is an exponential method of order 4 with two
         embedded methods of order 3 and 2 for an automatic step size
         selection. A dense output formula is also available.
         The computation of ((exp(A)-1)/A)*v, where A is the Jacobian
         matrix of RHSFunc and v is a vector is done by Krylov techniques.

         Theory, implementation issues and numerical examples can be
         found in: Marlis Hochbruck, Christian Lubich, Hubert Selhofer:
         "Exponential Integrators for Large Systems of Differential
         Equations", preprint. To have a look at it:

         http://na.uni-tuebingen.de/uni/mmn/eng/html/preprints.html

         To get the latest version of the code via ftp:

         ftp na.uni-tuebingen.de in directory /pub/codes/

 NOTE:   The system has to be *autonomous*, i.e. t never explicitly appears
         in RHSFunc. However, this is *NO* restriction, any problem can
         easily be reformulated as an autonomous one:
         Say, your righthand side function RHSFunc looks like
              function out = RHSFunc(t,y)
                out(1) = something in t,y(1),...,y(N);
                   ...
                out(N) = something in t,y(1),...,y(N);
         Introduce now the new component:
                out(N+1) = 1;
         (i.e.: the last component of the solution equals the time t)
         and add Y0(N+1) = T0 to your initial value. Replace t by
         Y(N+1) ind RHSFunc.

 USAGE:  See the examples described in the Makefile.
 (C)     Copyright (c) 1-12-96 by Marlis Hochbruck & Hubert Selhofer
         Marlis Hochbruck (marlis@na.uni-tuebingen.de)
         Hubert Selhofer (hubert@na.uni-tuebingen.de)
         University of Tuebingen, Germany.

 If you are interested in the further development of the code
 and a proper maintenance, then please send wishes, comments,
 experiences, suggestions and bug reports to the addresses above!
 Thank you!

 C code written by Mathias Froehlich

*/

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <setjmp.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"
#include "work.h"
#include "debug.h"

#include "errstep.h"
#include "h_init.h"
#include "output.h"
#include "scnorm.h"
#include "pfuns.h"
#include "ucalls.h"
#include "hbounds.h"

#ifdef EXP4EX
#include "phimge.h"
#include "phi_ex.h"
#else
#include "krystep.h"
#include "krylov.h"
#include "phikry.h"

#include "history.h"
#endif


#define FACMIN_KRY  1.0/3.0
#define GAMMA       1.0/3.0

static finteger inc1=1;

static RealCoef one  = RealCoefConst( 1.0);
static RealCoef zero = RealCoefConst( 0.0);
static RealCoef mone = RealCoefConst(-1.0);
#ifdef EXP4EX
static Complex wone  = ComplexConst( 1.0);
static Complex wzero = ComplexConst( 0.0);
#endif

/* for error exits from exp4 */
static jmp_buf Exp4ErrorExitJmpBuf;


#ifndef EXP4EX
int PREC(Exp4) (Real *t,Real tEnd,Complex *y,
      finteger neq,PREC(RhsFunc) RHSFunc,PREC(NewJac) jac,PREC(EvalJac) evjac,
      Exp4Boolean ConstantJacobian)
#else
int PREC(Exp4Ex) (Real *t,Real tEnd,Complex *y,
      finteger neq,PREC(RhsFunc) RHSFunc,PREC(FullJac) jac,
      Exp4Boolean ConstantJacobian)
#endif
     /* exp4 is a solver for big systems of stiff ODE's.

   Parameters:

   t          On entry it should containe the initial time of the problem.
              On exit it holds the time of the last calculated step.
              That is normally tEnd, if no error occured.

   tEnd       The end time you want to calculate.

   y          On entry y schould containe the initial values at time t.
              On exit it containes the solution vector an time t.

   neq        The number of equations, i.e. the dimension of Image(RHSFunc).

   RHSFunc    A pointer to the righthand side function of the ODE.

Exp4:
   jac        A pointer to the function which generates the jacobian of RHSFunc
              at the given point y, it has to be stored somwhere in your
	      module containing your function and jacobian evaluations.
	      The advantage of writing the jacobian into your module
	      is that you can use YOUR favorite sparse format.
              If it is set to the NULL pointer that function
	      is never referred.

   evjac      A pointer to a function which multiplies a vector with the
              jacobian. Or if jac==NULL it returns the derivative in
	      direction of the vector. If it is set to NULL exp4 takes
	      a routine for numerical differentiation to get that derivative.

FOR DETAILS SEE THE PROTOTYPES IN exp4.h!

Exp4Ex:
   jac        A pointer to the function which generates the jacobian of RHSFunc
              at the given point y. The result has to be written in the array
              res with the leading dimension ld.

FOR DETAILS SEE THE PROTOTYPES IN exp4.h!

Both:
   ConstantJacobian
              A flag which indicates if the jacobian is constant for all
              x in R^neq and t in R. ConstantJacobian=Exp4False if it is not
	      constant, else ConstantJacobian=Exp4True.

*/
{
  /* value returned by Exp4 */
  volatile int ReturnValue = 0;

  /* manage the exit at internal errors */
  if ((ReturnValue = setjmp(Exp4ErrorExitJmpBuf)) == 0)
  {
    /* for statistics */
    long AllSteps=0,FailSteps=0;
#ifndef EXP4EX
    long AllKry1=0,AllKryReuse=0,AllKry2=0,AllKry3=0;
#endif
    double Time;

    /* Workspace management */
    WorkSpace Work;

    /* data pointers */
    Complex *fy,*k__;
    /* help variable */
    RealCoef help;
    finteger ihelp;

    /* machine precision */
    const Real eps = (Real) plamch ("EPS");

    /* the actual stepsize */
    Real h;
    /* the last stepsize */
    Real h_old;
    Real h_old_unbounded;
    /* the approach to the new stepsize */
    Real h_new;

    /* error of the embedded W-method */
    Real errw = 1.0;
    /* error of the embedded order 3-method */
    Real err3;
    /* resulting error */
    Real err;

    /* flag: if true the jacobian will be reused */
    int reuse_J = 0;
    /* counts how long the Jacobian is reused */
    int reuse_J_count = 0;
    /* flag: will be true if last step is rejected */
    int reject = 0;
    /* flag: indicates if the jacobian is from an old step */
    int old_J=1;

#ifdef EXP4EX
    /* work variables */
    Complex *phia__,*a__;
#else
    /* struct for storing krylov stuff */
    krylov *Kry1,*Kry2,*Kry3;

    /* flag: indicates if the last step failed */
    int fail=0;
    /* flag: indicates if old k(1:3) will be reused */
    int reuse_k=0;

    History *KGuessHistory;
    History *Krylov1History,*Krylov2History,*Krylov3History;
#endif
    /* for the error stepsize control */
    Exp4ErrorStepSizeControlHandle ErrStepHandle = ErrstepHandleInitializer;


#define k(I)       (k__+((I)-1)*neq)
#define A(I,J)     a__[(I) + (J)*neq]
#define phiA(I,J)  phia__[(I) +(J)*neq]

    /* Some constants of the method */
    static const RealCoef
      alpha4[3] = { RealCoefConst(-7.0/3e2),  RealCoefConst(97.0/15e1),
		    RealCoefConst(-37.0/3e2)};
    static const RealCoef
      alpha7[7] = { RealCoefConst(59.0/3e2),  RealCoefConst(-7.0/75.0),
		    RealCoefConst(269.0/3e2), RealCoefConst(2.0/3.0),
		    RealCoefConst(2.0/3.0),   RealCoefConst(2.0/3.0)};
    static const RealCoef
      /* the coefs exp4 */
      alpha_exp4[7]={ RealCoefConst(0.0),     RealCoefConst(0.0),
		      RealCoefConst(1.0),     RealCoefConst(1.0),
		      RealCoefConst(-4.0/3.0),RealCoefConst(1.0),
		      RealCoefConst(1.0/6.0)};
    static const RealCoef
      /* the coefs for error of the w - method to the exp4 - method */
      alpha_errw[7]={ RealCoefConst(1.0),     RealCoefConst(-2.0),
		      RealCoefConst(1.0),     RealCoefConst(2.0),
		      RealCoefConst(-4.0/3.0),RealCoefConst(1.0),
		      RealCoefConst(-5.0/6.0)};
    static const RealCoef
      /* the coefs for error of the order 3 - method to the exp4 - method */
      alpha_err3[7]={ RealCoefConst(0.0),     RealCoefConst(0.0),
		      RealCoefConst(0.0),     RealCoefConst(3.0/2.0),
		      RealCoefConst(-2.0/3.0),RealCoefConst(1.0/2.0),
		      RealCoefConst(-1.0/3.0)};

    /* for error exit behavaour */
    Exp4SetJumpBuf(&Exp4ErrorExitJmpBuf);

    Exp4Message1((DEBUG_FILE,
	    "--------------------------------------------------------\n"
	    "        Exponential method of order 4(3,2) (W)\n"
	    "--------------------------------------------------------\n"));

    AllSteps = FailSteps = 0l;
    Time =  - F77CALL (dsecnd) ();

#ifdef EXP4EX
    {
      /* size for calculating Exp4PhimGe */
      size_t Size = PREC(SizePhimGeWork) (neq);
      /* size for the Jacobian, and for phi(Jacobian) */
      Size += 2*GetWorkSpaceSize(neq*neq*sizeof(Complex));
      /* size for k(1:7) */
      Size += GetWorkSpaceSize(sizeof(Complex)*neq*7);
      /* size for fy, d7,wu7,tmp (last three needed in the last stage) */
      Size += 4*GetWorkSpaceSize(sizeof(Complex)*neq);

      Work = Exp4GetWorkspace(Size);
      DistributeWorkspace(phia__,sizeof(Complex)*neq*neq,Work);
      DistributeWorkspace(a__,sizeof(Complex)*neq*neq,Work);
      DistributeWorkspace(fy,sizeof(Complex)*neq,Work);
      DistributeWorkspace(k__,sizeof(Complex)*neq*7,Work);
    }
#else  /* <=> #ifndef EXP4EX */
    PREC(Exp4GetKrylovSpaceHandles) (&Kry1,&Kry2,&Kry3,neq);
    /* calculating size of workspace for PhiKrylov */
    {
      /* Size for 1 vector */
      size_t SizeVec = GetWorkSpaceSize(sizeof(Complex)*neq);
      /* Size for k(1:3) */
      size_t SizeK13 = GetWorkSpaceSize(3*sizeof(Complex)*neq);
      /* Size for k(1:6) */
      size_t SizeK16 = GetWorkSpaceSize(6*sizeof(Complex)*neq);
      /* Size for k(1:7) */
      size_t SizeK17 = GetWorkSpaceSize(7*sizeof(Complex)*neq);
      /* workspace for the first krylov space */
      size_t SizeKry1 = PREC(Exp4SizePhiKryWork) (Kry1,3);
      /* workspace for the second krylov space */
      size_t SizeKry2 = PREC(Exp4SizePhiKryWork) (Kry2,3);
      /* workspace for the third krylov space */
      size_t SizeKry3 = PREC(Exp4SizePhiKryWork) (Kry3,1);
      /* Size for the first stage (included reuse_k) */
      size_t Stage1 = 2*SizeVec + 2*SizeK13 + SizeKry1;
      /* Size for the second stage */
      size_t Stage2 = 2*SizeVec + SizeK16 + max(SizeKry2,2*SizeVec);
      /* Size for the third stage */
      size_t Stage3 = 2*SizeVec + SizeK17 + max(SizeKry3,2*SizeVec);
      /* Size for the calculating errors */
      size_t Errors = SizeK17 + 2*SizeVec;
      /* final workspace size */
      size_t WorkSize =  max(Stage1,Stage2);
      WorkSize = max(WorkSize,Stage3);
      WorkSize = max(WorkSize,Errors);
      Work = Exp4GetWorkspace(WorkSize);

      DistributeWorkspace(fy,sizeof(Complex)*neq,Work);
      DistributeWorkspace(k__,sizeof(Complex)*neq*3,Work);
    }

    KGuessHistory = InitHistory(1);
    Krylov1History = InitHistory(3);
    Krylov2History = InitHistory(3);
    Krylov3History = InitHistory(3);
#endif
    PREC(Exp4InitOutput) (tEnd,1);

    /* tell the module hbounds about the end of integration time */
    PREC(Exp4SetTend) (tEnd);

    /* init module pfuns, it manages problem function calls without passing all
     function pointers to the subroutines of exp4. So the parameter lists are
     more readable */
    PREC(Exp4SetProblemFunctions) (Righthandside,RHSFunc);
#ifdef EXP4EX
    PREC(Exp4SetProblemFunctions) (FullJac,jac,ConstantJacobian);
#else
    PREC(Exp4SetProblemFunctions) (Jacobian,jac,ConstantJacobian);
    PREC(Exp4SetProblemFunctions) (Mv,evjac,neq);
#endif

    /* initialize the module for managing the user access of the inner steps */
    PREC(Exp4InitUcalls) (k__,neq,y,neq);

    /* calculating sc for the used scaled norm. Has to be done before
       Exp4GetInitalStepsize. */
    PREC(Exp4InitScNorm) (y,neq);

    /* doing the right hand side evaluation at the start point */
    PREC(Exp4EvalRhs) (neq,fy,y);

    /* get first step size h, k__ is here workspace */
    h_old_unbounded = PREC(Exp4GetInitialStepsize) (Work,fy,neq,y,4);
    h = PREC(Exp4ChkHBounds) (h_old_unbounded,*t);
    h_old = h*2.0;

    /* ---------------------------------------------------------------------
       The main loop
       --------------------------------------------------------------------- */

    while ((*t) < tEnd)
    {
      if (fabs((*t)*eps) > max(fabs(h),fabs(h_old)))
      {
	Warning("********* Detected Singularity ************");
	ReturnValue = EXP4_DETECT_SINGULARITY;
	break;
      }
#ifdef EXP4EX
      /* Do we need a new Jacobian? */
      if (!reuse_J)
      {
	reuse_J_count = 0;
	/* calculating the jacobian */
	if (old_J) PREC(Exp4GetNewFullJac) (neq,&A(0,0),neq,y),old_J=0;

	/* calculating phimA=(exp((h/3.0)*A)-Id)/((h/3.0)*A) */
	PREC(Exp4PhimGe)(Work,&phiA(0,0),neq,&A(0,0),neq,neq,h/3.0);
      } else {
	reuse_J_count++;
	/* If h changes then we need a new phiA */
	if (h_old != h)
	  PREC(Exp4PhimGe)(Work,&phiA(0,0),neq,&A(0,0),neq,neq,h/3.0);
      }

      /* calculate fy */
      if (!reject && (AllSteps != 0l)) PREC(Exp4EvalRhs) (neq,fy,y);

      /* get k(1:3) */
      PREC(phi_ex)(Work,k(1),&neq,3,fy,&neq,&A(0,0),&neq,&phiA(0,0),&neq,h);

#else /* <=> !defined(EXP4EX) */
      if (!reuse_k)
      {
	finteger KGuess;
	/* calculating new Jacobian */
	if ((!reuse_J || fail) && old_J) {
	  PREC(Exp4GetNewJac) (neq,y);
	  /* make schure that the Krylov-Supspace coresponds to
	     the actual Jacobian */ 
	  ClearKrylovSpace(Kry1);
	  old_J = 0;
	  reuse_J_count = 0;
	} else
	  reuse_J_count++;

	/* calculate fy */
	if (!reject && (AllSteps != 0l)) PREC(Exp4EvalRhs) (neq,fy,y);

	Exp4Message2((DEBUG_FILE,"Calculating new k(1:3)\n"));

	/* get the guessed krylov dimension (CGS uses it) */
	KGuess = GetMaximalHistoryEntry(KGuessHistory);

	/* if only one Krylov space is avaliable */
	if (!KrylovCanRebuild(Kry1)) ClearKrylovSpace(Kry1);
	fail = PREC(PhiKrylov)(Work,k(1),neq,Kry1,fy,&h,GAMMA,3,KGuess,
	      errw,1,neq);
	AllKry1 += KrylovSize(Kry1);
	AppendToHistory(KGuessHistory,KrylovKonvergence(Kry1));
      }
      else /* <=> if (reuse_k) */
      {
	Complex *fy_old,*k_new;
	reuse_J_count ++;
	DistributeWorkspace(k_new,3*neq*sizeof(Complex),Work);
	DistributeWorkspace(fy_old,neq*sizeof(Complex),Work);

	/* saving old fy into temp1 */
	pcopy (&neq,fy,&inc1,fy_old,&inc1);

	/* calculate fy */
	PREC (Exp4EvalRhs) (neq,fy,y);

	/* calculating fy_old-fy */
	praxpy (&neq,&mone,fy,&inc1,fy_old,&inc1);

	Exp4Message2((DEBUG_FILE,"Reusing k(1:3)\n"));

	/* Krylov space can not be reused because the start vector is new */
	ClearKrylovSpace(Kry1);

	/* calculating new k(1:3) = k_old(1:3) - phi(h*A)*(fy_old-fy) */
	fail = PREC(PhiKrylov)(Work,k_new,neq,Kry1,fy_old,&h,GAMMA,3,0,
	      errw,0,neq);
	if (KrylovSize(Kry1) > 0) {
	  ihelp = 3*neq;
	  praxpy (&ihelp,&mone,k_new,&inc1,k(1),&inc1);
	  AllKryReuse += KrylovSize(Kry1);
	}

	GiveBackWorkspace(fy_old,neq*sizeof(Complex),Work);
	GiveBackWorkspace(k_new,3*neq*sizeof(Complex),Work);
      } /* if (reuse_k) */

      ReDistributeWorkspace(k__,3*neq*sizeof(Complex),
	    6*neq*sizeof(Complex),Work);

      if (!fail)
      {
#endif /* !defined(EXP4EX) */
	{
	  Complex *d4,*wu4,*tmp;

	  DistributeWorkspace(d4,neq*sizeof(Complex),Work);
	  DistributeWorkspace(wu4,neq*sizeof(Complex),Work);
	  DistributeWorkspace(tmp,neq*sizeof(Complex),Work);
	  
	  /* w4*h = k(1:3) * (h*alpha4)'; (k(4) is used for w4*h) */
	  SetRealCoefVal(help,h);
	  ihelp = 3;
	  prgemv("N",&neq,&ihelp,&help,k(1),&neq,alpha4,&inc1,&zero,wu4,&inc1);
	  
	  /* calculate h*Jacobian*w4 into tmp */
#ifdef EXP4EX
	  pgemv ("No transpose",&neq,&neq,&wone,&A(0,0),&neq,wu4,&inc1,
		&wzero,tmp,&inc1);
#else
	  PREC(Exp4Mv) (neq,tmp,wu4,0);
#endif
	  /* calculate u4 = h*w4 + y into k(4) */
	  praxpy (&neq,&one,y,&inc1,wu4,&inc1);
	  
	  /* fu4 = f(u4) d4 holds fu4 */
	  PREC(Exp4EvalRhs) (neq,d4,wu4);
	  
	  /* add fy to h*Jacobian*w4 in tmp */
	  praxpy (&neq,&one,fy,&inc1,tmp,&inc1);
	  
	  /* d4 = fu4 - fy - h*Jacobian*w4 */
	  praxpy (&neq,&mone,tmp,&inc1,d4,&inc1);
	  
	  GiveBackWorkspace(tmp,neq*sizeof(Complex),Work);
	  GiveBackWorkspace(wu4,neq*sizeof(Complex),Work);
	  /* get k(4:6) */
#ifdef EXP4EX
	  PREC(phi_ex) (Work,k(4),&neq,3,d4,&neq,&A(0,0),&neq,
		&phiA(0,0),&neq,h);
#else /* defined(EXP4EX) */
	  ClearKrylovSpace(Kry2);

	  if (GetMaximalHistoryEntry(Krylov2History) <=
		KrylovPrimaryMaxSize(Kry2)) KrylovActivePrimary(Kry2);

	  /* k(4:6) = h*phi(h*Jacobian)*d4 */
	  fail = PREC(PhiKrylov)(Work,k(4),neq,Kry2,d4,&h,GAMMA,3,0,
		errw,0,neq);
	  AllKry2 += KrylovSize(Kry2);
#endif
	  GiveBackWorkspace(d4,neq*sizeof(Complex),Work);
	}
	
#ifndef EXP4EX
	ReDistributeWorkspace(k__,6*neq*sizeof(Complex),
	      7*neq*sizeof(Complex),Work);

	if (!fail)
	{
#endif
	  {
	    Complex *d7,*wu7,*tmp;

	    DistributeWorkspace(d7,neq*sizeof(Complex),Work);
	    DistributeWorkspace(wu7,neq*sizeof(Complex),Work);
	    DistributeWorkspace(tmp,neq*sizeof(Complex),Work);
	    
	    /* h*w7 = k(1:3)*(h*alpha7(1:3))'+(2*h/3)*(k(4)+k(5)+k(6)) */
	    /* k(7) is used for storing h*w7 */
	    SetRealCoefVal(help,h);
	    ihelp = 6;
	    prgemv ("N",&neq,&ihelp,&help,k(1),&neq,alpha7,&inc1,
		  &zero,wu7,&inc1);
	    
	    /* calculate h*Jacobian*w7 into temp2 */
#ifdef EXP4EX
	    pgemv ("No transpose",&neq,&neq,&wone,&A(0,0),&neq,wu7,&inc1,
		  &wzero,tmp,&inc1);
#else
	    PREC(Exp4Mv) (neq,tmp,wu7,0);
#endif
	    /* calculate u7 = h*w7 + y into k(7) */
	    praxpy (&neq,&one,y,&inc1,wu7,&inc1);
	    
	    /* fu7 = f(u7) d7 holds fu7 */
	    PREC(Exp4EvalRhs) (neq,d7,wu7);
	    
	    /* add fy to h*Jacobian*w7 in tmp */
	    praxpy (&neq,&one,fy,&inc1,tmp,&inc1);
	    
	    /* d7 = fu7 - fy - h*Jacobian*w7 */
	    praxpy (&neq,&mone,tmp,&inc1,d7,&inc1);
	    
	    GiveBackWorkspace(tmp,neq*sizeof(Complex),Work);
	    GiveBackWorkspace(wu7,neq*sizeof(Complex),Work);
	    /* get k(7) */
#ifdef EXP4EX
	    PREC(phi_ex) (Work,k(7),&neq,1,d7,&neq,&A(0,0),&neq,
		  &phiA(0,0),&neq,h);
#else
	    ClearKrylovSpace(Kry3);

	    if (GetMaximalHistoryEntry(Krylov3History) <=
		  KrylovPrimaryMaxSize(Kry3)) KrylovActivePrimary(Kry3);

	    fail=PREC(PhiKrylov) (Work,k(7),neq,Kry3,d7,&h,GAMMA,1,0,
		  errw,0,neq);
	    AllKry3 += KrylovSize(Kry3);
#endif
	    GiveBackWorkspace(d7,neq*sizeof(Complex),Work);
	  }

#ifndef EXP4EX
	  if (!fail)
#endif
	  {
#ifndef EXP4EX
	    Real h_err,h_kry;
#endif
	    Complex *y_new,*tmp;
	    DistributeWorkspace(y_new,neq*sizeof(Complex),Work);
	    DistributeWorkspace(tmp,neq*sizeof(Complex),Work);
	    /*******************************************/
	    /* calculating diff,diffw,diff3 and errors */
	    /*******************************************/
	    
	    /* y_new = y + h*(k(3)+k(4)-(4/3)*k(5)+k(6)+(1/6)*k(7)); */
	    pcopy (&neq,y,&inc1,y_new,&inc1);
	    SetRealCoefVal(help,h);
	    ihelp = 5;
	    prgemv ("N",&neq,&ihelp,&help,k(3),&neq,alpha_exp4+2,&inc1,
		  &one,y_new,&inc1);
	    
	    /* calculating sc */
	    PREC(Exp4SetScale) (y,y_new);
	    
	    /* calculating the errors of the embedded methods */
	    
	    /* get scaled norm of the embedded w - method */
	    ihelp = 7;
	    prgemv ("N",&neq,&ihelp,&one,k(1),&neq,alpha_errw,&inc1,
		  &zero,tmp,&inc1);
	    errw = fabs(h) * PREC(Exp4ScaledNorm) (tmp);
	    
	    /* get scaled norm of the embedded order 3 method */
	    ihelp = 4;
	    prgemv ("N",&neq,&ihelp,&one,k(4),&neq,alpha_err3+3,&inc1,
		  &zero,tmp,&inc1);
	    err3 = fabs(h) * PREC(Exp4ScaledNorm) (tmp);
	    
	    GiveBackWorkspace(tmp,neq*sizeof(Complex),Work);

	    err = min(errw,err3);
	    
	    if (err == 0.0) err = eps; /* for linear problems... */
	    
	    /* stepsize control */
#ifdef EXP4EX
	    /* Error estimation via Gustafson approach */
	    /* check h_min and h_max etc ... */
	    h_new = PREC(ErrstepGetNewH) (&ErrStepHandle,err,h,3);
#else
	    /* Error estimation via Gustafson approach */
	    h_err = PREC(ErrstepGetNewH) (&ErrStepHandle,err,h,3);

	    /* stepsize which might be good for the krylov process */
	    GetHKry(h_kry,h,Kry1,Krylov1History,Kry2,Krylov2History,
		  Kry3,Krylov3History);
	    
	    /* check h_min and h_max etc ... */
	    h_new = min(h_err,h_kry);
#endif
	    

#ifdef EXP4EX
	    Exp4Message1((DEBUG_FILE,
		  "t = %16.10e, h = %10.4e, errw = %10.4e, "
		  "err3 = %10.4e\n",(double)(*t),(double)h,(double)errw,
		  (double)err3));
#else
	    Exp4Message1((DEBUG_FILE,
		  "t = %12.6e, h=%8.2e, errw=%8.2e, err3=%8.2e "
		  "Krylov: %3d,%2d,%2d (%1s)\n",(double)(*t),(double)h,
		  (double)errw,(double)err3,(int)KrylovSize(Kry1),
		  (int)KrylovSize(Kry2),(int)KrylovSize(Kry3),
		  (h_kry<h_err)?"K":"E"));
#endif
	    /* -------------------------------------------------------------
	     * Update the solution only if the error is acceptable
	     * ------------------------------------------------------------- */
	    if (err <= 1.0)
	    {
	      AllSteps++;
	      
	      old_J = 1;
	      
	      /* reports the stepsize module the accepted errors and stepsize*/
	      PREC(ErrstepSetAccepted) (&ErrStepHandle,err,h);
#ifndef EXP4EX
	      if (reuse_k) {
		CopyLastEntry(Krylov1History);
	      } else {
		AppendToHistory(Krylov1History,KrylovSize(Kry1));
	      }
	      AppendToHistory(Krylov2History,KrylovSize(Kry2));
	      AppendToHistory(Krylov3History,KrylovSize(Kry3));
	      ClearKrylovSpace(Kry1);
#endif
	      {
		int outret;
		/* write output */
		outret = PREC(Exp4DoOutput)(Work,*t,h,y,y_new,k(1),neq);

		/* update the solution */
		(*t) += h;
		pcopy (&neq,y_new,&inc1,y,&inc1);

		if (outret) break;
	      }
	      
	      /* if the last stepsize was cut by the user take the stepsize
	       * approach of the step before for the new stepsize */
	      if (PREC(Exp4ChkHIsUpperBounded)()) h_new = h_old_unbounded;
	      h_old_unbounded = h_new;
	      
	      if (reject) reject = 0, h_new = min(h,h_new);
	      
	      h_old = h;
	      h = PREC(Exp4ChkHBounds) (h_new,*t);
	    }
	    else /* if (err <= 1.0) */
	    {
	      h_old = h;
	      h = PREC(Exp4ChkHBounds) (min(h,h_new),*t);
	      Exp4Message1((DEBUG_FILE,"== STEP REJECTED == STEP REJECTED "
		    "== STEP REJECTED ==\n"));

	      reject = 1;
	      FailSteps++;
	    }
	    
	    GiveBackWorkspace(y_new,neq*sizeof(Complex),Work);

#ifdef EXP4EX
	    reuse_J =((((h_new <= 2.0*h_old) && (err3 < 2.0*errw) &&
		  (errw <= 1.0)) || ConstantJacobian) &&
		  PREC(Exp4AllowReuse)());
#else
	    reuse_J = ((((h_kry <= h_err) && (errw<1.0) && (jac != NULL) &&
		  (GetHistoryEntry(Krylov1History,0)
			>= (2*max(KrylovSize(Kry2),KrylovSize(Kry3)))))
		  || ConstantJacobian) && PREC(Exp4AllowReuse)());

#endif
	    reuse_J = reuse_J
	      && (reuse_J_count < PREC(Exp4GetJacobianLifeTime) ());
	  } /* if (!fail) */
#ifndef EXP4EX
	} /* if (!fail) */

	ReDistributeWorkspace(k__,7*neq*sizeof(Complex),
	      6*neq*sizeof(Complex),Work);
      } /* if (!fail) */

      ReDistributeWorkspace(k__,6*neq*sizeof(Complex),
	    3*neq*sizeof(Complex),Work);

      if (fail)
      {
	Exp4Message1((DEBUG_FILE,
	      "== STEP FAILED == STEP FAILED == STEP FAILED ==\n"));
	reject = 1;
	h_new  = PREC(Exp4ChkHBounds) (FACMIN_KRY*h,*t);

	FailSteps++;

	h_old = h;
	h = min(h,h_new);
      }

      /* this krylov space can never be reused because the old fy is
       * overwritten by the new one. In case of failed or rejected steps
       * the k are never reused, so this is a krylov space for the wrong
       * start vector */
      if (reuse_k) {
	ClearKrylovSpace(Kry1);
	reuse_k = (reuse_J && (h == h_old) && !reject)
	  && (KrylovSize(Kry1) < GetHistoryEntry(Krylov1History,0));
      } else {
	reuse_k = (reuse_J && (h == h_old) && !reject);
      }
#endif
    } /* main loop while */

    Time += F77CALL (dsecnd) ();

#ifdef EXP4EX
    Exp4Message1((DEBUG_FILE,
	  "t = %16.10e\n\n"
	  "number of successful steps:          %11d\n"
	  "number of rejected steps:            %11d\n"
	  "number of fnc evaluations:           %11d\n"
	  "number of jacobi evaluations:        %11d\n"
	  "CPU time:\t\t\t\t%g seconds\n"
	  "%g <= Atol <= %g, %g <= Rtol <= %g\n\n"
	  "Total amount of allocated memory: %g %cBytes.\n\n\n",
	  (double)(*t),(int)AllSteps,(int)FailSteps,(int)PREC(FunEvals)(),
	  (int)PREC(JacEvals)(),(double)Time,
	  (double)PREC(Exp4GetMinAtol)(),
	  (double)PREC(Exp4GetMaxAtol)(),
	  (double)PREC(Exp4GetMinRtol)(),
	  (double)PREC(Exp4GetMaxRtol)(),
	  AmountOfMemory(), AmountOfMemoryUnit()));
#else
    Exp4Message1((DEBUG_FILE,
	  "t = %16.10e\n\n"
	  "number of successful steps:           %11d\n"
	  "number of rejected steps:             %11d\n"
	  "number of fnc evaluations:            %11d\n"
	  "number of jacobi evaluations:         %11d\n"
	  "number of Krylov steps 1st stage:     %11d\n"
	  "number of Krylov steps reuse k stage: %11d\n"
	  "number of Krylov steps 2nd stage:     %11d\n"
	  "number of Krylov steps 3rd stage:     %11d\n"
	  "number of all Krylov steps:           %11d\n"
	  "number of jacobian vector mult.:      %11d\n"
	  "number of transp. jac. vector mult.:  %11d\n\n"
	  "CPU time:\t\t\t\t%g seconds\n"
	  "%g <= Atol <= %g, %g <= Rtol <= %g\n\n"
	  "Total amount of allocated memory: %g %cBytes.\n\n\n",
	  (double)(*t),(int)AllSteps,(int)FailSteps,(int)PREC(FunEvals)(),
	  (int)PREC(JacEvals)(),(int)AllKry1,(int)AllKryReuse,(int)AllKry2,
	  (int)AllKry3,(int)(AllKry1+AllKryReuse+AllKry2+AllKry3),
	  (int)PREC(MvEvals)(),(int)PREC(vMEvals)(),(double)Time,
	  (double)PREC(Exp4GetMinAtol)(),
	  (double)PREC(Exp4GetMaxAtol)(),
	  (double)PREC(Exp4GetMinRtol)(),
	  (double)PREC(Exp4GetMaxRtol)(),
	  AmountOfMemory(), AmountOfMemoryUnit()));
#endif
  }
  Exp4FreeAllWorkspace();

  fflush(stdout);
  fflush(stderr);

  /* tell the error handler that the actual jmp_buff is invalid */
  Exp4SetJumpBuf(NULL);
  return ReturnValue;
}
