/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/
/*
 * This code is part of EXP4.
 *
 * This module handles the different ways build krylov spaces.
 */
#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"
#include "debug.h"

#include "pfuns.h"
#include "work.h"

#include "krylov.h"
#include "arnoldi.h"
#include "lanczos.h"
#include "phie1.h"
#include "krystep.h"

static finteger MaxKrylovDim = DEFAULT_MAX_KRYLOV_DIM;
static finteger MaxBlockSize = DEFAULT_MAX_BLOCK_SIZE;
static finteger MaxMixSize = DEFAULT_MAX_MIX_SIZE;
static Exp4KrylovStrategy Strategy;

void PREC(Exp4SetKrylovStrategy) (Exp4KrylovStrategy StrategyIn,...)
{
  va_list args;

  va_start(args,StrategyIn);

  switch (StrategyIn) {
  case Arnoldi:
  case ArnoldiOneSpace:
  case BiCGS:
  case BiConjugateGradientSqared:
  case BiCGSMixed:
  case BiConjugateGradientSqaredMixed:
  case BiCGSSwitched:
  case BiConjugateGradientSqaredSwitched:
    Strategy = StrategyIn;
    break;
  case Lanczos:
  case LanczosMixed:
  case LanczosSwitched:
    Strategy = StrategyIn;
    PREC(Exp4SetProblemFunctions) (Mtv,va_arg(args,PREC(EvalJacTransp)));
    break;
  default:
    Warning("No valid Exp4KrylovStrategy set, Exp4KrylovStrategy unchanged!");
    break;
  }

  va_end(args);
}

void PREC(Exp4SetMaxKrylovDim) (int dim)
{
  MaxKrylovDim = max(dim,1);
}

void PREC(Exp4SetMaxLanczosBlockSize) (int size)
{
  MaxBlockSize = max(size,1);
}

void PREC(Exp4SetMaxArnoldi23Size) (int size)
{
  MaxMixSize = max(1,size);
}

void PREC(Exp4GetKrylovSpaceHandles) (struct Krylov **Kry1,
      struct Krylov **Kry2,struct Krylov **Kry3,finteger neq)
{
  switch (Strategy) {
  case Arnoldi:
    Exp4Message2((DEBUG_FILE,
	  "Taking Arnoldi for all steps, maximum dimension = %d\n",
	  MaxKrylovDim));
    
    (*Kry1) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));

    break;
  case ArnoldiOneSpace:
    Exp4Message2((DEBUG_FILE,
	  "Taking Arnoldi for all stages, max dim = %d\n. Calc. all in the "
	  "same memory\n",MaxKrylovDim));

    (*Kry1) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Arn,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));

    break;
  case Lanczos:
    Exp4Message2((DEBUG_FILE,"Taking Look - Ahead Lanczos for all stages, "
	  "maximum dimension = %d\n",MaxKrylovDim));

    (*Kry1) = PREC(Exp4GetKrylovHandle) (Lan,MaxKrylovDim,MaxBlockSize,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Lan,MaxKrylovDim,
	  MaxBlockSize,neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Lan,MaxKrylovDim,
	  MaxBlockSize,neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    break;
  case LanczosMixed:
    {
      int UsedMaxMixSize = max(MaxMixSize,2*(2+MaxBlockSize));

      Exp4Message2((DEBUG_FILE,"Taking Lanczos for stage 1, max. dim. = "
	    "%d, max. blocks. %d\n",MaxKrylovDim,MaxBlockSize));
      Exp4Message2((DEBUG_FILE,"Taking Arnoldi for stage 2,3, maximum "
	    "dimension = %d, %d\n",UsedMaxMixSize+1,UsedMaxMixSize));

      (*Kry1) = PREC(Exp4GetKrylovHandle) (Lan,MaxKrylovDim,MaxBlockSize,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Arn,UsedMaxMixSize+1,0,
	    neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Arn,UsedMaxMixSize,0,
	    neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      break;
    }
  case LanczosSwitched:
    Exp4Message2((DEBUG_FILE,"Taking Lanczos for stage 1, max. dim. = "
	  "%d, max. blocks. %d\n",MaxKrylovDim,MaxBlockSize));
    Exp4Message2((DEBUG_FILE,"Taking Switched Lanczos - Arnoldi for stage 2,3 "
	  "max. dim = %d\n",MaxKrylovDim));
    
    (*Kry1) = PREC(Exp4GetKrylovHandle) (Lan,MaxKrylovDim,MaxBlockSize,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (LanSwitch,MaxKrylovDim,
	  MaxBlockSize,neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (LanSwitch,MaxKrylovDim,
	  MaxBlockSize,neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    break;
  case BiCGS:
  case BiConjugateGradientSqared:
    Exp4Message2((DEBUG_FILE,"Taking Cgs for all stages, maximum dimension "
	  "= %d\n",MaxKrylovDim));

    (*Kry1) = PREC(Exp4GetKrylovHandle) (Cgs,MaxKrylovDim,0,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Cgs,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Cgs,MaxKrylovDim,0,neq,0,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    break;
  case BiCGSMixed:
  case BiConjugateGradientSqaredMixed:
    {
      int UsedMaxMixSize = max(MaxMixSize,8);
      
      Exp4Message2((DEBUG_FILE,"Taking CGS for stage 1, max. dim. = %d\n",
	    MaxKrylovDim));
      Exp4Message2((DEBUG_FILE,"Taking Arnoldi for stage 2,3, maximum "
	    "dimension = %d, %d\n",UsedMaxMixSize+1,UsedMaxMixSize));
      
      (*Kry1) = PREC(Exp4GetKrylovHandle) (Cgs,MaxKrylovDim,0,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (Arn,UsedMaxMixSize+1,0,
	    neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (Arn,UsedMaxMixSize,0,
	    neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
      break;
    }
  case BiCGSSwitched:
  case BiConjugateGradientSqaredSwitched:
    Exp4Message2((DEBUG_FILE,"Taking CGS for stage 1, max. dim. = %d\n",
	  MaxKrylovDim));
    Exp4Message2((DEBUG_FILE,"Taking Switched Cgs - Arnoldi for stage 2,3 "
	  "max. dim = %d\n",MaxKrylovDim));
    
    (*Kry1) = PREC(Exp4GetKrylovHandle) (Cgs,MaxKrylovDim,0,neq,1,
	  &PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry2) (*Kry2) = PREC(Exp4GetKrylovHandle) (CgsSwitch,MaxKrylovDim,0,
	  neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    if (Kry3) (*Kry3) = PREC(Exp4GetKrylovHandle) (CgsSwitch,MaxKrylovDim,0,
	  neq,0,&PREC(Exp4PhiZ),&PREC(PhiZSizeWorkFun));
    break;
  default:
    ErrorReturn("Invalid Krylov space type set!",EXP4_ERR_INVALID_KRY_TYPE);
  }
}


/* creates a struct Krylov which holds data for a
 * Krylov function approximation */
struct Krylov *PREC(Exp4GetKrylovHandle) (Exp4KrylovType Type,
      finteger MaxKrylovDim,finteger MaxBlockSize,finteger neq,int CanRebuild,
      PhiZFunType PhiZFun,PhiZSizeWorkFunType PhiZSizeWorkFun)
{
  static int index[]={TEST_SEQUENCE};
  struct Krylov *Kry = Exp4GetWorkPointer(sizeof(struct Krylov));

  /* krylov type independant values */
  Kry->PhiZSizeWorkFun    = PhiZSizeWorkFun;
  Kry->PhiZFun            = PhiZFun;
  KrylovType(Kry)         = Type;
  KrylovNeq(Kry)          = neq;
  KrylovLdSmall(Kry)      = MaxKrylovDim;

  KrylovIndex(Kry)        = index;
  KrylovIndexLen(Kry)     = sizeof(index)/sizeof(int);

  KrylovLowStep(Kry)      = 4;
  KrylovGoodStep(Kry)     = PREC(GetNearestIndex) (4.0/7.0*MaxKrylovDim,
	KrylovIndex(Kry),KrylovIndexLen(Kry));
  KrylovOptStep(Kry)      = PREC(GetNearestIndex) (3.0/4.0*MaxKrylovDim,
	KrylovIndex(Kry),KrylovIndexLen(Kry));
  if (Type == Arn)
    KrylovExponent(Kry)   = 1.0/3.0;
  else
    KrylovExponent(Kry)   = 1.0/2.0;

  KrylovCanRebuild(Kry) = CanRebuild;

  ClearKrylovSpace(Kry);
  KrylovActivePrimary(Kry);

  switch (KrylovType(Kry)) {
  case Arn:
    PREC(Exp4FillHandleWithArnoldi) (Kry,MaxKrylovDim,neq,CanRebuild);
    break;
  case Lan:
    PREC(Exp4FillHandleWithLanczos)
      (Kry,MaxKrylovDim,MaxBlockSize,neq,CanRebuild);
    break;
  case LanSwitch:
    KrylovActiveSecondary(Kry);
    PREC(Exp4FillHandleWithLanczos)
      (Kry,MaxKrylovDim,MaxBlockSize,neq,CanRebuild);
    KrylovActivePrimary(Kry);
    PREC(Exp4FillHandleWithArnoldi) (Kry,MaxKrylovDim,neq,CanRebuild);
    break;
  case Cgs:
    PREC(Exp4FillHandleWithCgs) (Kry,MaxKrylovDim,neq,CanRebuild);
    break;
  case CgsSwitch:
    KrylovActiveSecondary(Kry);
    PREC(Exp4FillHandleWithCgs) (Kry,MaxKrylovDim,neq,CanRebuild);
    KrylovActivePrimary(Kry);
    PREC(Exp4FillHandleWithArnoldi) (Kry,MaxKrylovDim,neq,CanRebuild);
    break;
  }

  if (CanRebuild)
    KrylovH(Kry) = Exp4GetWorkPointer(sizeof(Complex)*
	  KrylovLdSmall(Kry)*KrylovMaxSize(Kry));

  return Kry;
}

