/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdio.h>
#include <stdarg.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "prec.h"
#include "work.h"
#include "errors.h"
#include "debug.h"

#include "exp4.h"
#include "output.h"


static finteger inc1=1;
static RealCoef one = RealCoefConst(1.0);

static Exp4OutputType Type = NoSpecial;
static PREC(SolTrait) SoloutFunction;
static FILE *File;
static Real *TSpan;
static finteger NTSpan;
static Complex *YOut;
static finteger LdYout;
static long StepNumber;
static Real tEnd;
static finteger NextY;
static finteger Next;

static struct DenseInfoStruct
{
  finteger neq;
  Complex *k;
  Real t;
  Real h;
  Complex *y_new;
  Complex *y_old;
} DenseInfoData;

static struct DenseInfoStruct *DenseInfo;

static int PREC(Exp4OutFileVector) (Complex *y,Real t,
      finteger neq);

static int PREC(Exp4OutFileDense) (WorkSpace Work,Complex *k,
      finteger neq,Complex *y_old,Complex *y_new,Real t,
      Real h);

static int PREC(Exp4OutDense) (Complex *k,finteger neq,
      Complex *y_old,Complex *y_new,Real t,
      Real h);

static void PREC(Exp4CalcDenseOutput) (Complex *output,finteger incout,
      Real tout,Complex *k,finteger neq,Complex *y_old,
      Complex *y_new,Real t,Real h);

static void PREC(Exp4OutGetPolynomial) (RealCoef *p,Real s,
      Real h);

void PREC(Exp4SetOutput) (Exp4OutputType TypeIn, ...)
{
  va_list list;

  va_start(list,TypeIn);
  Type = TypeIn;

  switch (Type) {
  case NoSpecial:
    break;
  case Dense:
    YOut   = va_arg(list,Complex*);
    LdYout = va_arg(list,finteger);
    TSpan  = va_arg(list,Real*);
    NTSpan = va_arg(list,finteger);
    break;
  case AllFile:
    File   = va_arg(list,FILE*);
    break;
  case DenseFile:
    File   = va_arg(list,FILE*);
    TSpan  = va_arg(list,Real*);
    NTSpan = va_arg(list,finteger);
    break;
  case EndFile:
    File   = va_arg(list,FILE*);
    break;
  case Solout:
    SoloutFunction = va_arg(list,PREC(SolTrait));
    break;
  default:
    Warning("Exp4SetOutput Warning: Illegal Type Value!");
    va_end(list);
  }

  va_end(list);
}

void PREC(Exp4InitOutput) (Real tEndIn,int EnableDense)
{
  StepNumber=1;
  NextY=0;
  Next=0;
  tEnd=tEndIn;

  if (EnableDense)
    DenseInfo = &DenseInfoData;
  else
  {
    DenseInfo = NULL;
    if ((Type == Dense) || (Type == DenseFile))
      ErrorReturn("requested dense output is not supported by this integrator",
	    EXP4_ERR_NO_DENSE);
  }
}

int PREC(Exp4DoOutput) (WorkSpace Work,Real t,Real h,
      Complex *y_old,Complex *y_new,Complex *k,finteger neq)
{
  /* write output */
  finteger ihelp=0;

  switch(Type)
  {
  case NoSpecial:
    return 0;
  case Dense:
    ihelp = PREC(Exp4OutDense) (k,neq,y_old,y_new,t,h);
    break;
  case AllFile:
    ihelp=PREC(Exp4OutFileVector)(y_new,t+h,neq);
    break;
  case DenseFile:
    ihelp =  PREC(Exp4OutFileDense) (Work,k,neq,y_old,y_new,t,h);
    break;
  case EndFile:
    if (t == tEnd-h)
      ihelp=PREC(Exp4OutFileVector)(y_new,t+h,neq);
    break;
  case Solout:
    {
      int solret=0;
#if (DEBUG >= 1)
      fflush(stdout);
#endif
      fflush(stderr);
      if (DenseInfo != NULL)
      {
	/* setting parameters for Dense output called from solout */
	DenseInfo->k = k;
	DenseInfo->neq = neq;
	DenseInfo->t = t;
	DenseInfo->h = h;
	DenseInfo->y_new = y_new;
	DenseInfo->y_old = y_old;
      }
      (*SoloutFunction) (StepNumber++,t,t+h,y_new,neq,&solret);
      if (solret<0)
	return 1;
      else
	return 0;
    }
  default:
    return 1;
  }

  if (ihelp != 0)
    ErrorReturn("Error in output routines!",EXP4_ERR_OUTPUT);

  return 0;
}

static int PREC(Exp4OutFileVector) (Complex *y,Real t,
      finteger neq)
{
  finteger i;

  if (ftell(File)<MAX_FILE_SIZE)
  {
    fprintf(File,"%.25e    ",t);
    for (i=0;i<neq;i++)
    {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
      fprintf(File,"%.25e  ",y[i]);
#else
      fprintf(File,"%.25e+%.25ei  ",y[i].r,y[i].i);
#endif
    }
    fprintf(File,"\n");
    return 0;
  }
  else
    ReturnMessage("MAX_FILE_SIZE exceeded!\nChange macro MAX_FILE_SIZE "
	  "in exp4.h!",1);
}

static int PREC(Exp4OutFileDense) (WorkSpace Work,Complex *k,
      finteger neq,Complex *y_old,Complex *y_new,Real t,
      Real h)
{
  Complex *Tmp;
  DistributeWorkspace(Tmp,sizeof(Complex)*neq,Work);

  while ((TSpan[Next]<=(t+h)) && Next<NTSpan)
  {
    PREC(Exp4CalcDenseOutput) (Tmp,1,TSpan[Next],k,neq,y_old,y_new,t,h);
    if (PREC(Exp4OutFileVector) (Tmp,TSpan[Next],neq)) return 1;
    Next++;
  }
  return 0;
}

static int PREC(Exp4OutDense) (Complex *k,finteger neq,
      Complex *y_old,Complex *y_new,Real t,
      Real h)
{
  while ((TSpan[Next]<=(t+h)) && Next<NTSpan)
  {
    PREC(Exp4CalcDenseOutput) (YOut+NextY,1,TSpan[Next],k,neq,y_old,y_new,t,h);
    NextY += LdYout;
    Next++;
  }
  return 0;
}

int PREC(Exp4GetDenseSolution) (Complex *output,finteger incout,
      Real tout)
{
  if (DenseInfo == NULL)
  {
      Warning("Dense output is not supported by this integrator\n");
      return 1;
  }
  else
  {
    if ((tout < DenseInfo->t) || (DenseInfo->t+DenseInfo->h < tout))
    {
      Warning("requested output time is not in the current integration "
	    "intervall\n");
      return 1;
    }
    else
    {
      PREC(Exp4CalcDenseOutput) (output,incout,tout,DenseInfo->k,
	    DenseInfo->neq,DenseInfo->y_old,DenseInfo->y_new,DenseInfo->t,
	    DenseInfo->h);
    }
    return 0;
  }
}

static void PREC(Exp4CalcDenseOutput) (Complex *output,finteger incout,
      Real tout,Complex *k,finteger neq,Complex *y_old,
      Complex *y_new,Real t,Real h)
{
  static finteger n=7;
  RealCoef p[7];
      
  if (tout == (t+h))
    pcopy (&neq,y_new,&inc1,output,&incout);
  else
  {
    PREC(Exp4OutGetPolynomial) (p,(tout-t)/h,h);
    
    pcopy (&neq,y_old,&inc1,output,&incout);
    prgemv("No transpose",&neq,&n,&one,k,&neq,p,&inc1,&one,output,&incout);
  }
}

static void PREC(Exp4OutGetPolynomial) (RealCoef *p, Real s, Real h)
{
  Real s2 = s*s;
  Real s3 = s2*s;
  Real s4 = s2*s2;

  SetRealCoefVal(p[0],(s-7.0/2.0*s2+5.0/3.0*s3)*h);
  SetRealCoefVal(p[1],2.0*s2*(1.0-s)*h);
  SetRealCoefVal(p[2],s3*h);
  SetRealCoefVal(p[3],s2*(2.0-4.0/3.0*s+1.0/3.0*s2)*h);
  SetRealCoefVal(p[4],-4.0/3.0*s4*h);
  SetRealCoefVal(p[5],s4*h);
  SetRealCoefVal(p[6],s2*(2.0/3.0*s-1.0/2.0)*h);
#if ((defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION)) || \
      !defined(USE_SECURE_COMPLEX_ARITMETIC))
  p[0]+=p[3]+p[4]+p[5]+p[6];
#else
  p[0].r+=p[3].r+p[4].r+p[5].r+p[6].r;
#endif
}
