/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdarg.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"
#include "debug.h"

#include "pfuns.h"

static PREC(RhsFunc)     RhsFun;
static PREC(FullJac)     FullJacobianFun;
static PREC(NewJac)      NewJacobianFun;
static PREC(EvalJac)     MvFun;
static PREC(EvalJacTransp) vMFun;

static Exp4Boolean ConstantJacobian;

static long nMv;
static long nvM;
static long FunEvals;
static long JacEvals;

static Complex *yJac;
static Real NormJac = 0.0;

void PREC(Exp4SetProblemFunctions) (Exp4FunctionType Type,...)
{
  va_list args;

  va_start(args,Type);
  switch (Type) {
  case Righthandside:
    RhsFun = va_arg(args,PREC(RhsFunc));
    FunEvals = 0l;
    if (RhsFun == NULL)
    {
      va_end(args);
      ErrorReturn("Missing functions for the problem!",EXP4_ERR_NO_RHS);
    }
    break;
  case Jacobian:
    NewJacobianFun   = va_arg(args,PREC(NewJac));
    ConstantJacobian = va_arg(args,Exp4Boolean);
    yJac = NULL;
    NormJac = 0.0;
    JacEvals = 0l;
    break;
  case FullJac:
    FullJacobianFun  = va_arg(args,PREC(FullJac));
    ConstantJacobian = va_arg(args,Exp4Boolean);
    JacEvals = 0l;
    if (FullJacobianFun == NULL)
    {
      va_end(args);
      ErrorReturn("Missing functions for the problem!",EXP4_ERR_NO_JAC);
    }
    break;
  case Mv:
    MvFun  = va_arg(args,PREC(EvalJac));
    nMv = 0l;
    if (MvFun == NULL)
    {
      va_end(args);
      ErrorReturn("Missing functions for the problem!",EXP4_ERR_NO_JAC);
    }
    break;
  case Mtv:
    vMFun  = va_arg(args,PREC(EvalJacTransp));
    nvM = 0l;
    break;
  }
  va_end(args);
}

void PREC(Exp4EvalRhs) (finteger neq,Complex *fy,Complex *y)
{
  Exp4Message4((DEBUG_FILE,"Calculating new righthandside.\n"));
  RhsFun(neq,fy,y);
  FunEvals++;
}

void PREC(Exp4GetNewFullJac) (finteger neq,Complex *a,finteger ld,
      Complex *y)
{
  if ((JacEvals == 0) || !ConstantJacobian)
  {
    Exp4Message2((DEBUG_FILE,"Calculating new jacobian.\n"));
    FullJacobianFun(neq,a,ld,y);
    JacEvals++;
  }
}

void PREC(Exp4GetNewJac) (finteger neq,Complex *y)
{
  if ((JacEvals == 0) || !ConstantJacobian)
  {
    if (NewJacobianFun != NULL)
    {
      Exp4Message2((DEBUG_FILE,"Calculating new jacobian.\n"));
      NewJacobianFun(neq,y);
      JacEvals++;
    }
    else
      yJac = y;
  }
  if (!ConstantJacobian)
    NormJac /= 2.0;
}

void PREC(Exp4Mv) (finteger neq,Complex *Mv,Complex *v,
      int normed_v)
{
  MvFun(neq,Mv,yJac,v);
  nMv++;
}

void PREC(Exp4vM) (finteger neq,Complex *vM,Complex *v)
{
  if (vMFun == NULL)
  {
    ErrorReturn("Missing functions for the problem!",EXP4_ERR_NO_VM);
  }
  else
  {
    vMFun(neq,vM,yJac,v);
    nvM++;
  }
}

void PREC(Exp4SetNormJacApproach) (Real Norm)
{
  /* multiply by a 'security' factor because all approaches given by
   * exp4 itself are too small all the time */
  Norm *= 1.5;
  NormJac = max(NormJac,Norm);

  Exp4Message9((DEBUG_FILE,"Approach to norm of the Jacobian: %e\n",NormJac));

}

Real PREC(Exp4GetNormJacApproach) (void)
{
  return NormJac;
}

long PREC(FunEvals) (void)
{
  return FunEvals;
}

long PREC(JacEvals) (void)
{
  return JacEvals;
}

long PREC(MvEvals) (void)
{
  return nMv;
}

long PREC(vMEvals) (void)
{
  return nvM;
}
