/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 * Matrix evaluation of the function
 *    phi(z) = (exp(z) - 1) / z
 * via Pade approximation or diagonalization.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 50 */
#define LOCAL_DEBUG_OFFSET 10
#include "debug.h"

#include "phi.h"

#include "phimge.h"


#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
#define PORMQR_TRANSPOSE "Transpose"
#else
#define PORMQR_TRANSPOSE "Conjugate transpose"
#endif

#define USE_QR_DECOMP_IN_PADE
/* #define USE_QR_DECOMP_IN_DIAG */
/* #define USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM */

#define PADE_N       6

/* the pade coeff's */
static RealCoef pade_n[PADE_N] = {
  RealCoefConst(1.0/26.0),RealCoefConst(5.0/156.0),
  RealCoefConst(1.0/858.0),RealCoefConst(1.0/5720.0),
  RealCoefConst(1.0/205920.0),RealCoefConst(1.0/8648640.0)};
static RealCoef pade_d[PADE_N] = {
  RealCoefConst(-6.0/13.0),RealCoefConst(5.0/52.0),
  RealCoefConst(-5.0/429.0),RealCoefConst(1.0/1144.0),
  RealCoefConst(-1.0/25740.0),RealCoefConst(1.0/1235520.0)};

#define MAX_COND     1E6

#define MAX_FULL     11

/* only short for INT_MAX, done because the table below is more pretty */
#define IMX  INT_MAX

static int max_dim_ge[12] = {IMX,100,60, 46,23,19,17,14,11,9,7 , 0};

static finteger inc1=1;
static finteger inc0=0;

static Complex wone  = ComplexConst(1.0);
static Real tone  = 1.0;
static Complex wzero = ComplexConst(0.0);

size_t PREC(SizePhimGeWork) (finteger roco)
/* returns the workspace size needed by phim */
{
  size_t DiagSize = PREC(SizePhimGeDiagWork) (roco);
  size_t PadeSize = PREC(SizePhimGePadeWork) (roco);

  return max(DiagSize,PadeSize);
}

int PREC(Exp4PhimGe) (WorkSpace Work,Complex *e,finteger lde,
      Complex *a,finteger lda,finteger roco,Real fac)

/*
  Exp4PhimGe calculates E=(exp(fac*A)-Id)/(fac*A) of a matrix A.

  Parameters:

  work      Is a pointer to the workspace which is needed to calculate E.
            max(5*roco*max(roco,3)*sizeof(Complex)+roco*sizeof(finteger),
            (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
	    2*roco*sizeof(finteger)) workspace required.

  e         Is a pointer to the result matrix E.

  lde       Is the leading dimension of E

  a         Is apointer to the input matrix A. unchanged on exit.

  lda       Is the leading dimension of A.

  roco      Is the dimension of A and E.

  fac       Is a real factor.
*/

{
  finteger result=0;

#define A(I,J) a[(I) + (J)*lda]
#define E(I,J) e[(I) + (J)*lde]

  if (roco == 1)
  {
    Complex z = A(0,0);
    rmul (z,fac,z);
    E(0,0) = PREC(phi) (z);
  }
  else
  {
    Real *LocalTemp;
    int s;
    Real normA;

    ShareWorkspace(LocalTemp,sizeof(Complex)*roco,Work);

    /* calculates the inf-norm of A */
    normA = plange ("Inf",&roco,&roco,&A(0,0),&lda,LocalTemp);

    Exp4Message2((DEBUG_FILE,"|| H || = %e\n",normA));

    /* get the scaling s */
    frexp(normA*fac,&s);
    s = max(0,s+1);

    /* If s is too big take the diagonalization routine.
       If the transformation to diagonal form is ill conditioned,
       result is equal 0, then use the pade approximation anyway. */
    if (roco>max_dim_ge[min(s,MAX_FULL)])
      result=PREC(Exp4PhimGeDiag) (Work,e,lde,a,lda,roco,fac);
    if (result==0)
      result=PREC(Exp4PhimGePade) (Work,e,lde,a,lda,roco,s,fac);
  }

  return result;
#undef A
#undef E
}


#ifndef USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM
size_t PREC(SizePhimGeDiagWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  finteger lwork=ld*max(4,ld);

  size_t WorkSize = GetWorkSpaceSize(sizeof(Complex)*ld*ld);

#ifdef USE_QR_DECOMP_IN_DIAG
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*ld);
#endif
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*ld);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*ld);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*ld);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*ld);
#else
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*ld);
  WorkSize += GetWorkSpaceSize(2*sizeof(Real)*ld);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*ld);
#endif
  return WorkSize;
}

int PREC(Exp4PhimGeDiag) (WorkSpace Work,Complex *e,finteger lde,
      Complex *a_org,finteger lda,finteger roco,Real fac)

     /*
       Calculates E=(exp(fac*A)-Id)/(fac*A) by diagonalization.
       It is called by Exp4PhimGe.

       Parameters:

       mem       Is a pointer to the workspace.
                 (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
                 2*roco*sizeof(finteger) bytes are required.

       e         Pointer to the result matrix E.

       lde       Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A.

       roco      Dimension of A and E.

       fac       Real factor.
       */

{
  Complex *a,*b,*work;
#ifdef USE_QR_DECOMP_IN_DIAG
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *tau;
#else
  Complex *tau;
#endif
#endif
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *evre,*evim;
#else
  Complex *eval;
  Real *rwork;
#endif

#define E(I,J)     e[(I) + (J)*lde]
#define A_org(I,J) a_org[(I) + (J)*lda]
#define A(I,J)     a[(I) + (J)*ld]
#define B(I,J)     b[(I) + (J)*ld]

  Complex help;
  finteger ld=roco;
  finteger *perm;
  finteger lwork=ld*max(4,ld);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  finteger *iwork;
#endif
  finteger info;
  finteger i;

  a=e;  /* memory management ... */

  DistributeWorkspace(b,sizeof(Complex)*ld*ld,Work);
#ifdef USE_QR_DECOMP_IN_DIAG
  DistributeWorkspace(tau,sizeof(Complex)*ld,Work);
#endif
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  DistributeWorkspace(evre,sizeof(Complex)*ld,Work);
  DistributeWorkspace(evim,sizeof(Complex)*ld,Work);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);
  DistributeWorkspace(iwork,sizeof(finteger)*ld,Work);
  DistributeWorkspace(perm,sizeof(finteger)*ld,Work);
#else
  DistributeWorkspace(eval,sizeof(Complex)*ld,Work);
  DistributeWorkspace(rwork,2*sizeof(Real)*ld,Work);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);
  DistributeWorkspace(perm,sizeof(finteger)*ld,Work);
#endif

  /* copying A_org because A_org is possibly needed to calculate via
     Pade approximation */
  placpy ("Full",&roco,&roco,&A_org(0,0),&lda,&A(0,0),&ld);

  /* getting the eigenvalues and eigenvectors of A, eigenvectors are in B */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  pgeev ("N","V",&roco,&A(0,0),&ld,evre,evim,NULL,&ld,&B(0,0),&ld,
	work,&lwork,&info);
#else
  pgeev ("N","V",&roco,&A(0,0),&ld,eval,NULL,&ld,&B(0,0),&ld,
	work,&lwork,rwork,&info);
#endif
  lapack_err(info,"Exp4PhimGeDiag","geev");

#if (DEBUG >= LOCAL_DEBUG_OFFSET+1)
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
#define eval evre
#endif
  {
    Real help = Re(eval[0]);
    for (i=1;i<roco;i++)
      if (help<Re(eval[i]))
	help = Re(eval[i]);
    Exp4Message2((DEBUG_FILE,"The largest realpart of the eigenvalues is: %e\n"
	  ,help*fac));
  }
#endif

  /*
   * copying the eigenvalue matrix,
   * E will be multiplied by phi(z), z the eigenvalues
   */
  placpy ("Full",&roco,&roco,&B(0,0),&ld,&E(0,0),&lde);

  {
    Real rcond;

#ifdef USE_QR_DECOMP_IN_DIAG
    for (i=0;i<roco;perm[i++]=1);
    pgeqpf (&roco,&roco,&B(0,0),&ld,perm,tau,work,&info);
    lapack_err(info,"Exp4PhimGeDiag","geqpf");

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    ptrcon ("Inf","Upper","No Unit",&roco,&B(0,0),&ld,&rcond,work,
	  iwork,&info);
#else
    ptrcon ("Inf","Upper","No Unit",&roco,&B(0,0),&ld,&rcond,work,
	  rwork,&info);
#endif
#else /* case to use LU - decomposition */
    /* calculating the inf-norm of B, pgecon needs that value */
    Real norm = plange ("Inf",&roco,&roco,&B(0,0),&ld,
	  (Real*)work);

    /* computes the LU factors of B for getting the condition number of A
       and for calculating .../B  */
    pgetrf (&roco,&roco,&B(0,0),&ld,&perm[0],&info);
    lapack_err(info,"Exp4PhimGe","getrf");

    /* gets the reciprocal condition number of B */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    pgecon ("Inf",&roco,&B(0,0),&ld,&norm,&rcond,work,iwork,&info);
#else
    pgecon ("Inf",&roco,&B(0,0),&ld,&norm,&rcond,work,
	  (Real*)(work+2*roco),&info);
#endif
    lapack_err(info,"Exp4PhimGe","gecon");
#endif

    Exp4Message1((DEBUG_FILE,"The condition number of the matrix of "
	  "eigenvalues: %e\n",1./rcond));

    if (1.0 < MAX_COND * rcond)
    {
      Exp4Message1((DEBUG_FILE,"Used diagonalization for calculating "
	    "Exp4PhimGe(Z)=(exp(Z)-1)/Z\n"));
      
      /* multiplying the eigenvalues, that means A, with fac */
      if (fac!=1.0)
      {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	pscal (&roco,&fac,evre,&inc1);
	pscal (&roco,&fac,evim,&inc1);
#else
	prscal (&roco,&fac,eval,&inc1);
#endif
      }

      for (i=0;i<roco;i++)
      {
	/* calculate A=B*diag((exp(eigenvalues)-1)/eigenvalues) */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	if ((*evim) == 0.0)
	{
	  /* if the eigenvalue i is real then calculate it with
	   * real arithmetic */
	  help = PREC(phi) (*evre);
	  
	  /* multiplying with the corresponding eigenvector */
	  pscal (&roco,&help,&E(0,i),&inc1);
	  
	  evre++; /* next eigenvalue */
	  evim++;
	}
	else
	{
	  /* if the eigenvalue i is complex calculate it with
	   * complex arithmetic */
#ifdef SINGLE_PRECISION
	  fsinglecomplex chelp;
	  chelp.r = *evre;
	  chelp.i = *evim;
	  chelp = Cphi (chelp);
#else
	  fdoublecomplex chelp;
	  chelp.r = *evre;
	  chelp.i = *evim;
	  chelp = Zphi (chelp);
#endif
	  /* multiplying with the corresponding eigenvectors */
	  prot (&roco,&E(0,i+1),&inc1,&E(0,i),&inc1,&chelp.r,&chelp.i);
	  
	  i++;     /* two eigenvalues are done in the imaginary case */
	  evre+=2;
	  evim+=2;
	}
#else
	help = PREC(phi) (*eval);
	
	/* multiplying with the corresponding eigenvector */
	pscal (&roco,&help,&E(0,i),&inc1);
	
	eval++; /* next eigenvalue */
#endif
      } /* for all eigenvalues */
      
      {
#ifdef USE_QR_DECOMP_IN_DIAG
	flogical flag = FTRUE;
	
	plapmt (&flag,&roco,&roco,&E(0,0),&lde,perm);
	
	ptrsm ("Right","Upper","No transpose","No Unit",&roco,&roco,
	      &wone,&B(0,0),&ld,&E(0,0),&lde);
	
	pormqr ("Right",PORMQR_TRANSPOSE,&roco,&roco,&roco,&B(0,0),&ld,tau,
	      &E(0,0),&lde,work,&lwork,&info);
	lapack_err(info,"Exp4PhimGeDiag","ormqr");
#else /* case to use LU - decomposition */
	/* solves the equation Res = A/B */
	ptrsm ("Right","Upper","No transpose","Non-unit",&roco,&roco,
	      &wone,&B(0,0),&ld,&E(0,0),&lde);
	
	ptrsm ("Right","Lower","No transpose","Unit",&roco,&roco,&wone,
	      &B(0,0),&ld,&E(0,0),&lde);
	
	/* reverse the row pivoting */
	for (i=roco-1;0<=i;i--)
	{
	  finteger ihelp = perm[i]-1;
	  if (i != ihelp)
	    pswap (&roco,&E(0,i),&inc1,&E(0,ihelp),&inc1);
	}
#endif
      }
      return 1;
    }
  }
  return 0;

#undef E
#undef A_org
#undef A
#undef B
}

#else

size_t PREC(SizePhimGeDiagWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  finteger lwork=ld*max(4,ld);

  size_t WorkSize = 2*GetWorkSpaceSize(sizeof(Complex)*roco*ld);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  WorkSize += 3*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*roco);
#else
  WorkSize += 2*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += 2*GetWorkSpaceSize(sizeof(Real)*2*roco);
#endif
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);
  return WorkSize;
}

int PREC(Exp4PhimGeDiag) (WorkSpace Work,Complex *PhiA,finteger LdPhiA,
      Complex *A,finteger LdA,finteger roco,Real fac)

     /*
       Calculates E=(exp(fac*A)-Id)/(fac*A) by diagonalization.
       It is called by Exp4PhimGe.

       Parameters:

       mem       Is a pointer to the workspace.
                 (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
                 2*roco*sizeof(finteger) bytes are required.

       e         Pointer to the result matrix E.

       lde       Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A.

       roco      Dimension of A and E.

       fac       Real factor.
       */

{
  Complex *Vl;
  Complex *Vr;
  Real *Scale;
  Complex *Tau;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *EvRe;
  Complex *EvIm;
#else
  Complex *Ev;
  Real *rwork;
#endif
  Complex *work;
  finteger ld=roco;
  finteger ilo,ihi;
  finteger nout;
  flogical select;
  finteger *iwork;
  finteger info,lwork;
  int i;

#define VR(I) (Vr+(I)*ld)
#define VL(I) (Vl+(I)*ld)


  DistributeWorkspace(Vl,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Vr,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Scale,sizeof(Real)*roco,Work);
  DistributeWorkspace(Tau,sizeof(Complex)*roco,Work);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  DistributeWorkspace(EvRe,sizeof(Complex)*roco,Work);
  DistributeWorkspace(EvIm,sizeof(Complex)*roco,Work);
#else
  DistributeWorkspace(Ev,sizeof(Complex)*roco,Work);
  DistributeWorkspace(rwork,sizeof(Real)*2*roco,Work);
#endif
  DistributeWorkspace(iwork,sizeof(finteger)*roco,Work);
  lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);

  /* copying A to PhiA */
  placpy ("Full",&roco,&roco,A,&LdA,PhiA,&ld);

  /* getting the eigenvalues and eigenvectors of A, eigenvectors are in B */
  /* this is adapted from dgeev.f to use the upper Hessenberg structure */

  /* balancing A */
  pgebal ("B",&roco,PhiA,&ld,&ilo,&ihi,Scale,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebal");

  /* reduce to upper hessenberg */
  pgehrd (&roco,&ilo,&ihi,PhiA,&ld,Tau,work,&lwork,&info);
  lapack_err(info,"Exp4PhimHsDiag","gehrd");

  placpy ("L",&roco,&roco,PhiA,&ld,Vl,&ld);
  
  porghr (&roco,&ilo,&ihi,Vl,&ld,Tau,work,&lwork,&info);
  lapack_err(info,"Exp4PhimHsDiag","orghr");

  /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,EvRe,EvIm,Vl,&ld,
	work,&lwork,&info);
#else
  phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,Ev,Vl,&ld,
	work,&lwork,&info);
#endif
  lapack_err(info,"Exp4PhimHsDiag","hseqr");

  placpy ("F",&roco,&roco,Vl,&ld,Vr,&ld);
  
  /* get the eigenvectors */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  ptrevc ("B","B",&select,&roco,PhiA,&ld,Vl,&ld,Vr,&ld,&roco,&nout,
	work,&info);
#else
  ptrevc ("B","B",&select,&roco,PhiA,&ld,Vl,&ld,Vr,&ld,&roco,&nout,
	work,rwork,&info);
#endif
  lapack_err(info,"Exp4PhimHsDiag","trevc");
    
  /* backtransform the left eigenvectors (balancing) */
  pgebak ("B","L",&roco,&ilo,&ihi,Scale,&roco,Vl,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");
  
  /* backtransform the right eigenvectors (balancing) */
  pgebak ("B","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");

  /* normalize the right eigenvalues to get the 1-norm of the matrix
   * of left eigenvectors is equal one. Then scale each right eigenvector
   * to get Vl'*Vr = Id. */
  for (i=0;i<roco;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    if (EvIm[i] == 0.0) {
      /* scaling to 1-norm of VL eq one */
      Real scl = 1.0/pasum(&roco,VL(i),&inc1);
      pscal (&roco,&scl,VL(i),&inc1);

      /* scaling for Vl'*Vr = Id */
      pdot (&scl,&roco,VR(i),&inc1,VL(i),&inc1);
      scl = 1.0/scl;
      pscal (&roco,&scl,VR(i),&inc1);
    } else {
      /* scaling to 1-norm of VL eq one */
      Real RealPart = pasum(&roco,VL(i),&inc1);
      Real ImagPart = pasum(&roco,VL(i+1),&inc1);
      Real help = 1.0/max(RealPart,ImagPart);
      pscal (&roco,&help,VL(i),&inc1);
      pscal (&roco,&help,VL(i+1),&inc1);

      /* scaling for Vl'*Vr = Id */
      pdot (&RealPart,&roco,VL(i),&inc1,VR(i),&inc1);
      pdot (&ImagPart,&roco,VL(i+1),&inc1,VR(i),&inc1);
      help = 1.0/(RealPart*RealPart+ImagPart*ImagPart);
      RealPart *= help; ImagPart *= help;
      prot (&roco,VR(i+1),&inc1,VR(i),&inc1,&RealPart,&ImagPart);

      i++;
    }
#else
    {
      /* scaling to 1-norm of VL eq one */
      Real scl = 1.0/pasum(&roco,VL(i),&inc1);
      prscal (&roco,&scl,VL(i),&inc1);
    }
    {
      Complex Dot;
      Real help;
      /* scaling for Vl'*Vr = Id */
      pdot (&Dot,&roco,VR(i),&inc1,VL(i),&inc1);
      help = 1.0/(Dot.r*Dot.r+Dot.i*Dot.i);
      Dot.r *= help; Dot.i *= help;
      pscal (&roco,&Dot,VR(i),&inc1);
    }
#endif
  }

  {
    /* this is the condition of the diagonaltransformation
     * (1-norm of Vl is equal one, that is the inf-norm of Vl' is equal one) */
    Real Cond = plange ("I",&roco,&roco,VR(0),&ld,PhiA);
    if (MAX_COND < Cond)
      return 0;
    else {
      for (i=0;i<roco;i++)
      {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	if (EvIm[i] == 0.0) {
	  /* multiplying with phi(Eigenvalue) */
	  Complex help;
	  help = PREC(phi) (EvRe[i]*fac);
	  pscal (&roco,&help,VL(i),&inc1);
	} else {
	  /* multiplying with phi(Eigenvalue) */
#ifdef SINGLE_PRECISION
	  fsinglecomplex chelp;
	  chelp.r = EvRe[i]*fac;
	  chelp.i = EvIm[i]*fac;
	  chelp = Cphi (chelp);
#else
	  fdoublecomplex chelp;
	  chelp.r = EvRe[i]*fac;
	  chelp.i = EvIm[i]*fac;
	  chelp = Zphi (chelp);
#endif
	  chelp.i = -chelp.i;
	  prot (&roco,VL(i+1),&inc1,VL(i),&inc1,&chelp.r,&chelp.i);

	  i++;
	}
#else
	/* multiplying with phi(Eigenvalue) */
	Complex help;
	help.r = fac*Ev[i].r;
	help.i = fac*Ev[i].i;
	help = PREC(phi) (help);
	conj(help,help);
	pscal (&roco,&help,VL(i),&inc1);
#endif
      }

      pgemm ("N","C",&roco,&roco,&roco,&wone,Vr,&ld,Vl,&ld,
	    &wzero,PhiA,&LdPhiA);

      return 1;
    }
  }
}
#endif

size_t PREC(SizePhimGePadeWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  int matsize = max(ld,3)*roco;

  size_t WorkSize = 5*GetWorkSpaceSize(sizeof(Complex)*matsize);

  WorkSize += GetWorkSpaceSize(sizeof(finteger)*ld);

  return WorkSize;
}

int PREC(Exp4PhimGePade) (WorkSpace Work,Complex *res,finteger ldres,
	  Complex *a_org,finteger lda,finteger roco,int s,
	  Real fac)

     /*
       calculates E=(exp(fac*A)-Id)/(fac*A) by Pade approximation.
       It is called by Exp4PhimGe.

       Parameters:

       mem       Is a pointer to the workspace.
                 5*roco*max(roco,3)*sizeof(Complex)+roco*sizeof(finteger)
                 bytes are required.

       res       Pointer to the result matrix E.

       ldres     Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A and E.

       roco      Dimension of A and E.

       s         Gives the scaling factor 2^s for getting the inf-norm of A
                 less equal then 1/2.

       fac       Real factor.
       */

{
  finteger i;
  finteger ld=roco;
  finteger incld1=ld+1;
  finteger ld2=ld*ld;
  finteger *perm,info;
  Complex *temp,*unused,*a,*d,*e,*expvar,*x;

#define A_org(I,J)   a_org[(I) + (J)*lda]
#define Res(I,J)     res[(I) + (J)*ldres]
#define A(I,J)       a[(I) + (J)*ld]
#define D(I,J)       d[(I) + (J)*ld]
#define E(I,J)       e[(I) + (J)*ld]
#define Exp(I,J)     expvar[(I) + (J)*ld]
#define Unused(I,J)  unused[(I) + (J)*ld]
#define X(I,J)       x[(I) + (J)*ld]

  ShareWorkspace(a,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(x,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(temp,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(d,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(e,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(unused,sizeof(Complex)*ld*max(ld,3),Work);
  DistributeWorkspace(perm,sizeof(finteger)*ld,Work);


  Exp4Message1((DEBUG_FILE,"Used pade app. for calc. "
	"Exp4PhimGe(Z)=(exp(Z)-1)/Z fac=%e s=%d\n",fac,s));

  /* the scaling */
  fac = ldexp(fac,-s);

  /* copying A_org because A_org is possibly needed  later times */
  placpy ("Full",&roco,&roco,&A_org(0,0),&lda,&A(0,0),&ld);

  /* multiplying A with fac */
  plascl ("G",&inc1,&inc1,&tone,&fac,&roco,&roco,&A(0,0),&ld,&info);
  lapack_err(info,"Exp4PhimGePade","lascl");


  /* Pade approximation for (exp(z)-1)/z */

  /* initialize iteration */

  /* copying A into D,E */
  placpy ("full",&roco,&roco,&A(0,0),&ld,&E(0,0),&ld);
  placpy ("full",&roco,&roco,&A(0,0),&ld,&D(0,0),&ld);

  /* multiplying E with pade_n[0] */
#if ((defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION)) || \
      !defined(USE_SECURE_COMPLEX_ARITMETIC))
  plascl ("G",&inc1,&inc1,&tone,&pade_n[0],&roco,&roco,&E(0,0),&ld,
	&info);
#else
  plascl ("G",&inc1,&inc1,&tone,&pade_n[0].r,&roco,&roco,&E(0,0),&ld,
	&info);
#endif
  lapack_err(info,"Exp4PhimGePade","lascl");

  /* multiplying D with pade_d[0] */
#if ((defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION)) || \
      !defined(USE_SECURE_COMPLEX_ARITMETIC))
  plascl ("G",&inc1,&inc1,&tone,&pade_d[0],&roco,&roco,&D(0,0),&ld,
	&info);
#else
  plascl ("G",&inc1,&inc1,&tone,&pade_d[0].r,&roco,&roco,&D(0,0),&ld,
	&info);
#endif
  lapack_err(info,"Exp4PhimGePade","lascl");

  /* adding the identity matrix to E and D */
  paxpy (&roco,&wone,&wone,&inc0,&E(0,0),&incld1);
  paxpy (&roco,&wone,&wone,&inc0,&D(0,0),&incld1);


  /* the Iteration */

  for (i=1;i<PADE_N;i++)
  {
    /* X=X*A */
    pgemm ("N","N",&roco,&roco,&roco,&wone,&X(0,0),&ld,&A(0,0),&ld,
	  &wzero,&Unused(0,0),&ld);

    if (i!=1)     /* if it is the first step &X(0,0) points to A */
      temp=x;    /* so new X is in Unused and temp points to an empty array. */
    x=unused;    /* (in the first step temp points to an empty array by */
    unused=temp; /* initialization in the other steps it points to that array
                    that becomes empty */

    /* E=E+pade_n[k]*X */
    praxpy (&ld2,&pade_n[i],&X(0,0),&inc1,&E(0,0),&inc1);

    /* D=D+pade_d[k]*X */
    praxpy (&ld2,&pade_d[i],&X(0,0),&inc1,&D(0,0),&inc1);
  } /* for */


  /* last step of Pade approximation: the right division */

#ifdef USE_QR_DECOMP_IN_PADE
  {
    Complex *tau=&X(0,0); /* array X is not used in the moment */
    finteger lwork = roco*roco;
    Complex *dwork = &Unused(0,0);
    flogical flag = FFALSE;

    for (i=0;i<roco;perm[i++]=0);
    pgeqpf (&roco,&roco,&D(0,0),&ld,perm,tau,dwork,&info);
    lapack_err(info,"Exp4PhimGePade","geqpf");

    pormqr ("Left",PORMQR_TRANSPOSE,&roco,&roco,&roco,&D(0,0),&ld,tau,
	  &E(0,0),&ld,dwork,&lwork,&info);
    lapack_err(info,"Exp4PhimGePade","ormqr");

    ptrtrs ("Upper","No transpose","No Unit",&roco,&roco,&D(0,0),&ld,
	  &E(0,0),&ld,&info);
    lapack_err(info,"Exp4PhimGePade","trtrs");

    for (i=0;i<roco;i++)
      plapmt (&flag,&inc1,&roco,&E(0,i),&inc1,perm);
  }
#else
  pgesv (&roco,&roco,&D(0,0),&ld,&perm[0],&E(0,0),&ld,&info);
  lapack_err(info,"Exp4PhimGePade","gesv");
#endif

  if (0<s)  /* Undo Scaling. */
  {
    Complex half = ComplexConst(0.5);

    /* calculates Exp=A*E , initialize the iteration */
    pgemm ("N","N",&roco,&roco,&roco,&wone,&A(0,0),&ld,&E(0,0),&ld,
	  &wzero,&Unused(0,0),&ld);

    expvar=unused;
    unused=d;   /* the values in D never used */

    /* adding the identity matrix to Exp */
    paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);


    for (i=1;i<s;i++)  /* The Iteration */
    {
      /* calculates E=0.5*(Exp+Id)*E */

      /* save the Diagonal elements */
      pcopy (&roco,&Exp(0,0),&incld1,&Res(0,0),&inc1);

      /* adds Id to Exp */
      paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);

      /* Unused=0.5*(Exp+id)*E */
      pgemm ("N","N",&roco,&roco,&roco,&half,&Exp(0,0),&ld,
	    &E(0,0),&ld,&wzero,&Unused(0,0),&ld);

      /* rebuild Exp */
      pcopy (&roco,&Res(0,0),&inc1,&Exp(0,0),&incld1);

      temp=e;  /* memory management .... */
      e=unused;
      unused=temp;

      /* calculates Exp=Exp*Exp */
      pgemm ("N","N",&roco,&roco,&roco,&wone,&Exp(0,0),&ld,
	    &Exp(0,0),&ld,&wzero,&Unused(0,0),&ld);

      temp=expvar;  /* memory management .... */
      expvar=unused;
      unused=temp;
    }       /* nearly end of iteration, */


    /*  the last step follows now: */


    /* calculates the return value Res=0.5*(Exp+Id)*E */

    /* adds id to Exp */
    paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);

    /* calculates Res=0.5*Exp*E */
    pgemm ("N","N",&roco,&roco,&roco,&half,&Exp(0,0),&ld,
	  &E(0,0),&ld,&wzero,&Res(0,0),&ldres);
  }
  else /* copying if no scaling was done */
    placpy ("Full",&roco,&roco,&E(0,0),&ld,&Res(0,0),&ldres);

  return 1;

#undef A_org
#undef Res
#undef A
#undef D
#undef E
#undef Exp
#undef Unused
#undef X
}
