/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

/*
 * Matrix evaluation of the function
 *    phi(z) = (exp(z) - 1) / z
 * via Pade approximation or diagonalization.
 */
#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <limits.h>
#include "ftypes.h"
#include "fblas.h"
#include "flapack.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"

/* #define LOCAL_DEBUG_LEVEL 50 */
#define LOCAL_DEBUG_OFFSET 10
#include "debug.h"

#include "phi.h"

#include "phimhs.h"
#include "phimge.h"


#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
#define PORMQR_TRANSPOSE "Transpose"
#else
#define PORMQR_TRANSPOSE "Conjugate transpose"
#endif

#define USE_QR_DECOMP_IN_PADE
#define USE_BAND_DECOMP
/* #define USE_QR_DECOMP_IN_DIAG */
/* #define USE_HESSENBERG_FORM */
/* #define USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM */

#define PADE_N       6

static void bandedmm(finteger roco,Complex *res,finteger ldres,
      Complex *a,finteger lda,finteger bandau,finteger bandal,
      Complex *b,finteger ldb,finteger bandbu,finteger bandbl);

/* the pade coeff's */
static RealCoef pade_n[PADE_N]={
  RealCoefConst(1.0/26.0),RealCoefConst(5.0/156.0),
  RealCoefConst(1.0/858.0),RealCoefConst(1.0/5720.0),
  RealCoefConst(1.0/205920.0),RealCoefConst(1.0/8648640.0)};
static RealCoef pade_d[PADE_N]={
  RealCoefConst(-6.0/13.0),RealCoefConst(5.0/52.0),
  RealCoefConst(-5.0/429.0),RealCoefConst(1.0/1144.0),
  RealCoefConst(-1.0/25740.0),RealCoefConst(1.0/1235520.0)};

#define MAX_COND     1E6

#define MAX_FULL     11
#define MAX_BAND     3

/* only short for INT_MAX, done because the table below is more pretty */
#define IMX  INT_MAX

static int max_dim_he[MAX_BAND+1][MAX_FULL+1] = {
/* s = 0   1   2   3   4   5   6   7   8   9   10 >11 */
     {IMX,IMX,IMX,IMX,IMX,100, 50, 30, 20, 17, 11,  0},  /* bandwith = 1 */
     {IMX,IMX,IMX,IMX,200,100, 46, 28, 20, 15, 11,  0},  /* bandwith = 2 */
     {IMX,IMX,500,300,100, 80, 40, 25, 18, 13, 10,  0},  /* bandwith = 3 */
     {IMX,200,100, 60, 46, 23, 17, 13, 11,  8,  7,  0}}; /* bandwith > 3 */

static int max_dim_ge[12]    =
   {IMX,100,60, 46,23,19,17,14,11,9,7 , 0};

static finteger inc1=1;
static finteger inc0=0;

static Complex wone  = ComplexConst(1.0);
static Complex wzero = ComplexConst(0.0);

size_t PREC(SizePhimHsWork) (finteger roco)
/* returns the workspace size needed by phim */
{
  size_t DiagSize = PREC(SizePhimHsDiagWork) (roco);
  size_t PadeGeSize = PREC(SizePhimGePadeWork) (min(roco,MAX_FULL));
  size_t PadeHeSize = PREC(SizePhimHsPadeWork) (roco);

  return max(max(DiagSize,PadeGeSize),PadeHeSize);
}

int PREC(Exp4PhimHs) (WorkSpace Work,Complex *e,finteger lde,
      Complex *a,finteger lda,finteger roco,finteger bandwith,
      Real fac)
/*
  Exp4PhimHs calculates E=(exp(fac*A)-Id)/(fac*A) of a matrix A.

  Parameters:

  work      Is a pointer to the workspace which is needed to calculate E.
            max(5*roco*max(roco,3)*sizeof(Complex)+roco*sizeof(finteger),
            (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
	    2*roco*sizeof(finteger)) workspace required.

  e         Is a pointer to the result matrix E.

  lde       Is the leading dimension of E

  a         Is apointer to the input matrix A. unchanged on exit.

  lda       Is the leading dimension of A.

  roco      Is the dimension of A and E.

  fac       Is a real factor.
*/

{
  finteger result=0;

#define A(I,J) a[(I) + (J)*lda]
#define E(I,J) e[(I) + (J)*lde]

  if (roco == 1)
  {
    Complex z = A(0,0);
    rmul (z,fac,z);
    E(0,0) = PREC(phi) (z);
  }
  else
  {
    Real *LocalTemp;
    int s;
    Real normA;
    finteger lda1 = lda+1;
    bandwith = min(bandwith,roco-1);

    ShareWorkspace(LocalTemp,sizeof(Complex)*roco,Work);

    /* calculates the inf-norm of A */
    normA = plangb ("Inf",&roco,&inc1,&bandwith,&A(-bandwith,0),&lda1,
	  LocalTemp);

    Exp4Message2((DEBUG_FILE,"|| H || = %e\n",normA));

    /* get the scaling s */
    frexp(normA*fac,&s);
    s=max(0,s+1);

    if (roco <= MAX_FULL)
    {
      /* If s is too big take the diagonalization routine.
	 If the transformation to diagonal form is ill conditioned,
	 result is equal 0, then use the pade approximation anyway. */
      if (roco>max_dim_ge[min(s,MAX_FULL)])
	result=PREC(Exp4PhimHsDiag)(Work,e,lde,a,lda,roco,fac);
      if (result==0)
	result=PREC(Exp4PhimGePade)(Work,e,lde,a,lda,roco,s,fac);
    }
    else
    {
      /* If s is too big take the diagonalization routine.
	 If the transformation to diagonal form is ill conditioned,
	 result is equal 0, then use the pade approximation anyway. */
      if (roco>max_dim_he[max(min(bandwith,MAX_BAND)-1,0)][min(s,MAX_FULL)])
	result=PREC(Exp4PhimHsDiag)(Work,e,lde,a,lda,roco,fac);
      if (result==0)
	result=PREC(Exp4PhimHsPade)(Work,e,lde,a,lda,roco,bandwith,s,fac);
    }
  }
  return result;
#undef A
#undef E
}

#ifdef USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM
size_t PREC(SizePhimHsDiagWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  finteger lwork=ld*max(4,ld);

  size_t WorkSize = 2*GetWorkSpaceSize(sizeof(Complex)*roco*ld);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  WorkSize += 3*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*roco);
#else
  WorkSize += 2*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*2*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*roco);
#endif
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);

  return WorkSize;
}

int PREC(Exp4PhimHsDiag) (WorkSpace Work,Complex *PhiA,finteger LdPhiA,
      Complex *A,finteger LdA,finteger roco,Real fac)

     /*
       Calculates E=(exp(fac*A)-Id)/(fac*A) by diagonalization.
       It is called by Exp4PhimGe.

       Parameters:

       mem       Is a pointer to the workspace.
                 (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
                 2*roco*sizeof(finteger) bytes are required.

       e         Pointer to the result matrix E.

       lde       Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A.

       roco      Dimension of A and E.

       fac       Real factor.
       */

{
  Complex *Vl;
  Complex *Vr;
  Real *Scale;
  Complex *Tau;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *EvRe;
  Complex *EvIm;
#else
  Complex *Ev;
  Real *rwork;
#endif
  Complex *work;
  finteger ld=roco;
  finteger ilo,ihi;
  finteger nout;
  flogical select;
  finteger *iwork;
  finteger info,lwork;
  int i;

#define VR(I) (Vr+(I)*ld)
#define VL(I) (Vl+(I)*ld)


  DistributeWorkspace(Vl,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Vr,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Scale,sizeof(Real)*roco,Work);
  DistributeWorkspace(Tau,sizeof(Complex)*roco,Work);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  DistributeWorkspace(EvRe,sizeof(Complex)*roco,Work);
  DistributeWorkspace(EvIm,sizeof(Complex)*roco,Work);
#else
  DistributeWorkspace(Ev,sizeof(Complex)*roco,Work);
  DistributeWorkspace(rwork,sizeof(Real)*2*roco,Work);
#endif
  DistributeWorkspace(iwork,sizeof(finteger)*roco,Work);
  lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);

  /* copying A to PhiA */
  placpy ("Full",&roco,&roco,A,&LdA,PhiA,&ld);

  /* getting the eigenvalues and eigenvectors of A, eigenvectors are in B */
  /* this is adapted from dgeev.f to use the upper Hessenberg structure */

#ifdef USE_HESSENBERG_FORM
  /* balancing A */
  pgebal ("S",&roco,PhiA,&ld,&ilo,&ihi,Scale,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebal");

  /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  phseqr("S","I",&roco,&ilo,&ihi,PhiA,&ld,EvRe,EvIm,Vl,&ld,
	work,&lwork,&info);
#else
  phseqr("S","I",&roco,&ilo,&ihi,PhiA,&ld,Ev,Vl,&ld,
	work,&lwork,&info);
#endif
  lapack_err(info,"Exp4PhimHsDiag","hseqr");
#else
  /* balancing A */
  pgebal ("B",&roco,PhiA,&ld,&ilo,&ihi,Scale,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebal");

  /* reduce to upper hessenberg */
  pgehrd (&roco,&ilo,&ihi,PhiA,&ld,Tau,work,&lwork,&info);
  lapack_err(info,"Exp4PhimHsDiag","gehrd");

  placpy ("L",&roco,&roco,PhiA,&ld,Vl,&ld);
  
  porghr (&roco,&ilo,&ihi,Vl,&ld,Tau,work,&lwork,&info);
  lapack_err(info,"Exp4PhimHsDiag","orghr");

  /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,EvRe,EvIm,Vl,&ld,
	work,&lwork,&info);
#else
  phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,Ev,Vl,&ld,
	work,&lwork,&info);
#endif
  lapack_err(info,"Exp4PhimHsDiag","hseqr");
#endif

  placpy ("F",&roco,&roco,Vl,&ld,Vr,&ld);
  
  /* get the eigenvectors */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  ptrevc ("B","B",&select,&roco,PhiA,&ld,Vl,&ld,Vr,&ld,&roco,&nout,
	work,&info);
#else
  ptrevc ("B","B",&select,&roco,PhiA,&ld,Vl,&ld,Vr,&ld,&roco,&nout,
	work,rwork,&info);
#endif
  lapack_err(info,"Exp4PhimHsDiag","trevc");
    
#ifdef USE_HESSENBERG_FORM
  /* backtransform the left eigenvectors (balancing) */
  pgebak ("S","L",&roco,&ilo,&ihi,Scale,&roco,Vl,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");
  
  /* backtransform the right eigenvectors (balancing) */
  pgebak ("S","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");
#else
  /* backtransform the left eigenvectors (balancing) */
  pgebak ("B","L",&roco,&ilo,&ihi,Scale,&roco,Vl,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");
  
  /* backtransform the right eigenvectors (balancing) */
  pgebak ("B","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
  lapack_err(info,"Exp4PhimHsDiag","gebak");
#endif

  /* normalize the right eigenvalues to get the 1-norm of the matrix
   * of left eigenvectors is equal one. Then scale each right eigenvector
   * to get Vl'*Vr = Id. */
  for (i=0;i<roco;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    if (EvIm[i] == 0.0) {
      /* scaling to 1-norm of VL eq one */
      Real scl = 1.0/pasum(&roco,VL(i),&inc1);
      pscal (&roco,&scl,VL(i),&inc1);

      /* scaling for Vl'*Vr = Id */
      pdot (&scl,&roco,VR(i),&inc1,VL(i),&inc1);
      scl = 1.0/scl;
      pscal (&roco,&scl,VR(i),&inc1);
    } else {
      /* scaling to 1-norm of VL eq one */
      Real RealPart = pasum(&roco,VL(i),&inc1);
      Real ImagPart = pasum(&roco,VL(i+1),&inc1);
      Real help = 1.0/max(RealPart,ImagPart);
      pscal (&roco,&help,VL(i),&inc1);
      pscal (&roco,&help,VL(i+1),&inc1);

      /* scaling for Vl'*Vr = Id */
      pdot (&RealPart,&roco,VL(i),&inc1,VR(i),&inc1);
      pdot (&ImagPart,&roco,VL(i+1),&inc1,VR(i),&inc1);
      help = 1.0/(RealPart*RealPart+ImagPart*ImagPart);
      RealPart *= help; ImagPart *= help;
      prot (&roco,VR(i+1),&inc1,VR(i),&inc1,&RealPart,&ImagPart);

      i++;
    }
#else
    {
      /* scaling to 1-norm of VL eq one */
      Real scl = 1.0/pasum(&roco,VL(i),&inc1);
      prscal (&roco,&scl,VL(i),&inc1);
    }
    {
      Complex Dot;
      Real help;
      /* scaling for Vl'*Vr = Id */
      pdot (&Dot,&roco,VR(i),&inc1,VL(i),&inc1);
      help = 1.0/(Dot.r*Dot.r+Dot.i*Dot.i);
      Dot.r *= help; Dot.i *= help;
      pscal (&roco,&Dot,VR(i),&inc1);
    }
#endif
  }

  {
    /* this is the condition of the diagonaltransformation
     * (1-norm of Vl is equal one, that is the inf-norm of Vl' is equal one) */
    Real Cond = plange ("I",&roco,&roco,VR(0),&ld,PhiA);
    if (MAX_COND < Cond)
      return 0;
    else {
      for (i=0;i<roco;i++)
      {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	if (EvIm[i] == 0.0) {
	  /* multiplying with phi(Eigenvalue) */
	  Complex help;
	  help = PREC(phi) (EvRe[i]*fac);
	  pscal (&roco,&help,VL(i),&inc1);
	} else {
	  /* multiplying with phi(Eigenvalue) */
#ifdef SINGLE_PRECISION
	  fsinglecomplex chelp;
	  chelp.r = EvRe[i]*fac;
	  chelp.i = EvIm[i]*fac;
	  chelp = Cphi (chelp);
#else
	  fdoublecomplex chelp;
	  chelp.r = EvRe[i]*fac;
	  chelp.i = EvIm[i]*fac;
	  chelp = Zphi (chelp);
#endif
	  chelp.i = -chelp.i;
	  prot (&roco,VL(i+1),&inc1,VL(i),&inc1,&chelp.r,&chelp.i);

	  i++;
	}
#else
	/* multiplying with phi(Eigenvalue) */
	Complex help;
	help.r = fac*Ev[i].r;
	help.i = fac*Ev[i].i;
	help = PREC(phi) (help);
	conj(help,help);
	pscal (&roco,&help,VL(i),&inc1);
#endif
      }

      pgemm ("N","C",&roco,&roco,&roco,&wone,Vr,&ld,Vl,&ld,
	    &wzero,PhiA,&LdPhiA);

      return 1;
    }
  }
}

#else /* USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM */

size_t PREC(SizePhimHsDiagWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  finteger lwork=ld*max(4,ld);

  size_t WorkSize = GetWorkSpaceSize(sizeof(Complex)*roco*ld);

  WorkSize += GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(Real)*roco);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  WorkSize += 2*GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += 2*GetWorkSpaceSize(sizeof(finteger)*roco);
#else
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*roco);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*2*roco);
  WorkSize += GetWorkSpaceSize(sizeof(finteger)*roco);
#endif
  WorkSize += GetWorkSpaceSize(sizeof(Complex)*lwork);
  return WorkSize;
}

int PREC(Exp4PhimHsDiag) (WorkSpace Work,Complex *PhiA,finteger LdPhiA,
      Complex *A,finteger LdA,finteger roco,Real fac)

     /*
       Calculates E=(exp(fac*A)-Id)/(fac*A) by diagonalization.
       It is called by Exp4PhimHs.

       Parameters:

       mem       Is a pointer to the workspace.
                 (roco*roco+roco*max(4,roco)+2*roco)*sizeof(Complex)+
                 2*roco*sizeof(finteger) bytes are required.

       e         Pointer to the result matrix E.

       lde       Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A.

       roco      Dimension of A and E.

       fac       Real factor.
       */

{
  Complex *Vr;
  Complex *work;
  Complex *Tau;
  Real *Scale;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  Complex *EvRe,*EvIm;
#else
  Complex *Ev;
  Real *rwork;
#endif
  Complex help;
  finteger ld=roco;
  finteger *perm;
  finteger lwork;
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  finteger *iwork;
#endif
  finteger info;
  finteger i;

#define VR(i) (Vr+(i)*ld)

  DistributeWorkspace(Vr,sizeof(Complex)*roco*ld,Work);
  DistributeWorkspace(Tau,sizeof(Complex)*roco,Work);
  DistributeWorkspace(Scale,sizeof(Real)*roco,Work);
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  DistributeWorkspace(EvRe,sizeof(Complex)*roco,Work);
  DistributeWorkspace(EvIm,sizeof(Complex)*roco,Work);
  DistributeWorkspace(iwork,sizeof(finteger)*roco,Work);
  DistributeWorkspace(perm,sizeof(finteger)*roco,Work);
#else
  DistributeWorkspace(Ev,sizeof(Complex)*roco,Work);
  DistributeWorkspace(rwork,2*sizeof(Real)*roco,Work);
  DistributeWorkspace(perm,sizeof(finteger)*roco,Work);
#endif
  lwork = GetRemainingWorkSpaceSize(Work)/sizeof(Complex);
  DistributeWorkspace(work,sizeof(Complex)*lwork,Work);

  /* copying A into PhiA Pade approximation */
  placpy ("Full",&roco,&roco,A,&LdA,PhiA,&ld);

  /* getting the eigenvalues and eigenvectors of A, eigenvectors are in Vr */
  /* this is adapted from dgeev.f to use the upper Hessenberg structure */
  {
    finteger ilo,ihi;
    finteger nout;
    flogical select;

#ifdef USE_HESSENBERG_FORM
    /* balancing A */
    pgebal ("S",&roco,PhiA,&ld,&ilo,&ihi,Scale,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebal");

    /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    phseqr("S","I",&roco,&ilo,&ihi,PhiA,&ld,EvRe,EvIm,Vr,&ld,
	  work,&lwork,&info);
#else
    phseqr("S","I",&roco,&ilo,&ihi,PhiA,&ld,Ev,Vr,&ld,
	  work,&lwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","hseqr");
#else /* USE_HESSENBERG_FORM */
    /* balancing A */
    pgebal ("B",&roco,PhiA,&ld,&ilo,&ihi,Scale,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebal");
    
    /* reduce to upper hessenberg */
    pgehrd (&roco,&ilo,&ihi,PhiA,&ld,Tau,work,&lwork,&info);
    lapack_err(info,"Exp4PhimHsDiag","gehrd");
    
    placpy ("L",&roco,&roco,PhiA,&ld,Vr,&ld);
    
    porghr (&roco,&ilo,&ihi,Vr,&ld,Tau,work,&lwork,&info);
    lapack_err(info,"Exp4PhimHsDiag","orghr");
    
    /* do QR itteration */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,EvRe,EvIm,Vr,&ld,
	  work,&lwork,&info);
#else
    phseqr("S","V",&roco,&ilo,&ihi,PhiA,&ld,Ev,Vr,&ld,
	  work,&lwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","hseqr");
#endif /* USE_HESSENBERG_FORM */

    /* get the eigenvectors */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    ptrevc ("R","B",&select,&roco,PhiA,&ld,NULL,&ld,Vr,&ld,&roco,&nout,
	  work,&info);
#else
    ptrevc ("R","B",&select,&roco,PhiA,&ld,NULL,&ld,Vr,&ld,&roco,&nout,
	  work,rwork,&info);
#endif
    lapack_err(info,"Exp4PhimHsDiag","trevc");
    
#ifdef USE_HESSENBERG_FORM
    /* backtransform the right eigenvectors (balancing) */
    pgebak ("S","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebak");
#else /* USE_HESSENBERG_FORM */
    /* backtransform the right eigenvectors (balancing) */
    pgebak ("B","R",&roco,&ilo,&ihi,Scale,&roco,Vr,&ld,&info);
    lapack_err(info,"Exp4PhimHsDiag","gebak");
#endif /* USE_HESSENBERG_FORM */

    /* Normalize eigenvectors and make largest component real */
    for (i=0;i<roco;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
      if (EvIm[i] == 0.0) {
	Real scl = 1.0/pnrm2(&roco,VR(i),&inc1);
	pscal (&roco,&scl,VR(i),&inc1);
      } else {
	int j;
	Real normi = pnrm2(&roco,VR(i),&inc1);
	Real normi1 = pnrm2(&roco,VR(i+1),&inc1);
	Real scl = 1.0/sqrt(normi*normi+normi1*normi1);
	pscal (&roco,&scl,VR(i),&inc1);
	pscal (&roco,&scl,VR(i+1),&inc1);
	for (j=0;j<roco;j++)
	{
	  Real help1 = *(VR(i)+j);
	  Real help2 = *(VR(i+1)+j);
	  work[j]=help2*help2+help1*help1;
	}
	j = ipamax(&roco,work,&inc1)-1;
	{
	  Real cs,sn,r;
	  plartg(VR(i)+j,VR(i+1)+j,&cs,&sn,&r);
	  prot(&roco,VR(i),&inc1,VR(i+1),&inc1,&cs,&sn);
	  (*(VR(i+1)+j)) = 0.0;
	}
	i++;
      }
#else
      Real scl = 1.0/pnrm2(&roco,VR(i),&inc1);
      prscal (&roco,&scl,VR(i),&inc1);
#endif
    }
  }

#if (DEBUG >= LOCAL_DEBUG_OFFSET+1)
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
  {
    Real help = EvRe[0];
    for (i=1;i<roco;i++)
      if (help<EvRe[i])
	help = EvRe[i];
    Exp4Message2((DEBUG_FILE,"The largest realpart of the eigenvalues is: %e\n"
	  ,help*fac));
  }
#else
  {
    Real help = Re(Ev[0]);
    for (i=1;i<roco;i++)
      if (help<Re(Ev[i]))
	help = Re(Ev[i]);
    Exp4Message2((DEBUG_FILE,"The largest realpart of the eigenvalues is: %e\n"
	  ,help*fac));
  }
#endif
#endif

  /*
   * copying the eigenvalue matrix,
   * E will be multiplied by phi(z), z the eigenvalues
   */
  placpy ("Full",&roco,&roco,Vr,&ld,PhiA,&LdPhiA);

  {
    Real rcond;

#ifdef USE_QR_DECOMP_IN_DIAG
    for (i=0;i<roco;perm[i++]=1);
    pgeqpf (&roco,&roco,Vr,&ld,perm,Tau,work,&info);
    lapack_err(info,"Exp4PhimHsDiag","geqpf");

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    ptrcon ("Inf","Upper","No Unit",&roco,Vr,&ld,&rcond,work,
	  iwork,&info);
#else
    ptrcon ("Inf","Upper","No Unit",&roco,Vr,&ld,&rcond,work,
	  rwork,&info);
#endif
#else /* case to use LU - decomposition */
    /* calculating the inf-norm of B, pgecon needs that value */
    Real norm = plange ("Inf",&roco,&roco,Vr,&ld,
	  (Real*)work);

    /* computes the LU factors of B for getting the condition number of A
       and for calculating .../B  */
    pgetrf (&roco,&roco,Vr,&ld,&perm[0],&info);
    lapack_err(info,"Exp4PhimHs","getrf");

    /* gets the reciprocal condition number of B */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    pgecon ("Inf",&roco,Vr,&ld,&norm,&rcond,work,iwork,&info);
#else
    pgecon ("Inf",&roco,Vr,&ld,&norm,&rcond,work,
	  (Real*)(work+2*roco),&info);
#endif
    lapack_err(info,"Exp4PhimHs","gecon");
#endif

    Exp4Message1((DEBUG_FILE,"The condition number of the matrix of "
	  "eigenvalues: %e\n",1./rcond));

    if (1.0 < MAX_COND * rcond)
    {
      Exp4Message1((DEBUG_FILE,"Used diagonalization for calculating "
	    "Exp4PhimHs(Z)=(exp(Z)-1)/Z\n"));
      
      /* multiplying the eigenvalues, that means A, with fac */
      if (fac!=1.0)
      {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	pscal (&roco,&fac,EvRe,&inc1);
	pscal (&roco,&fac,EvIm,&inc1);
#else
	prscal (&roco,&fac,Ev,&inc1);
#endif
      }

      for (i=0;i<roco;i++)
      {
	/* calculate A=B*diag((exp(eigenvalues)-1)/eigenvalues) */
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	if (EvIm[i] == 0.0)
	{
	  /* if the eigenvalue i is real then calculate it with
	   * real arithmetic */
	  help = PREC(phi) (EvRe[i]);
	  
	  /* multiplying with the corresponding eigenvector */
	  pscal (&roco,&help,PhiA+i*LdPhiA,&inc1);
	}
	else
	{
	  /* if the eigenvalue i is complex calculate it with
	   * complex arithmetic */
#ifdef SINGLE_PRECISION
	  fsinglecomplex chelp;
	  chelp.r = EvRe[i];
	  chelp.i = EvIm[i];
	  chelp = Cphi (chelp);
#else
	  fdoublecomplex chelp;
	  chelp.r = EvRe[i];
	  chelp.i = EvIm[i];
	  chelp = Zphi (chelp);
#endif
	  /* multiplying with the corresponding eigenvectors */
	  prot (&roco,PhiA+(i+1)*LdPhiA,&inc1,PhiA+i*LdPhiA,&inc1,
		&chelp.r,&chelp.i);
	  
	  i++;     /* two eigenvalues are done in the imaginary case */
	}
#else
	help = PREC(phi) (Ev[i]);
	
	/* multiplying with the corresponding eigenvector */
	pscal (&roco,&help,PhiA+i*LdPhiA,&inc1);
#endif
      } /* for all eigenvalues */
      
      {
#ifdef USE_QR_DECOMP_IN_DIAG
	flogical flag = FTRUE;
	
	plapmt (&flag,&roco,&roco,PhiA,&LdPhiA,perm);
	
	ptrsm ("Right","Upper","No transpose","No Unit",&roco,&roco,
	      &wone,Vr,&ld,PhiA,&LdPhiA);
	
	pormqr ("Right",PORMQR_TRANSPOSE,&roco,&roco,&roco,Vr,&ld,Tau,
	      PhiA,&LdPhiA,work,&lwork,&info);
	lapack_err(info,"Exp4PhimHsDiag","ormqr");
#else /* case to use LU - decomposition */
	Complex one = ComplexConst(1.0);
	
	/* solves the equation Res = A/B */
	ptrsm ("Right","Upper","No transpose","Non-unit",&roco,&roco,
	      &one,Vr,&ld,PhiA,&LdPhiA);
	
	ptrsm ("Right","Lower","No transpose","Unit",&roco,&roco,&one,
	      Vr,&ld,PhiA,&LdPhiA);
	
	/* reverse the row pivoting */
	for (i=roco-1;0<=i;i--)
	{
	  finteger ihelp = perm[i]-1;
	  if (i != ihelp)
	    pswap (&roco,PhiA+i*LdPhiA,&inc1,PhiA+ihelp*LdPhiA,&inc1);
	}
#endif
      }
      return 1;
    }
  }
  return 0;

#undef VR
}
#endif /* else USE_LEFT_EIGENVECTORS_FOR_BACKTRANSFORM */


size_t PREC(SizePhimHsPadeWork) (finteger roco)
     /* returns the workspace size needed by phim */
{
  finteger ld=roco;
  int matsize = max(ld,3)*roco;

  size_t WorkSize = 5*GetWorkSpaceSize(sizeof(Complex)*matsize);

  WorkSize += GetWorkSpaceSize(sizeof(finteger)*ld);

  return WorkSize;
}

int PREC(Exp4PhimHsPade) (WorkSpace Work,Complex *res,finteger ldres,
      Complex *a_org,finteger lda,finteger roco,finteger bandwith,
      int s,Real fac)

     /*
       calculates E=(exp(fac*A)-Id)/(fac*A) by Pade approximation.
       It is called by Exp4PhimHs.

       Parameters:

       mem       Is a pointer to the workspace.
                 5*roco*max(roco,3)*sizeof(Complex)+roco*sizeof(finteger)
                 bytes are required.

       res       Pointer to the result matrix E.

       ldres     Leading dimension of E.

       a_org     Pointer to the input matrix A. unchanged on exit.

       lda       Leading dimension of A and E.

       roco      Dimension of A and E.

       s         Gives the scaling factor 2^s for getting the inf-norm of A
                 less equal then 1/2.

       fac       Real factor.
       */

{
  finteger i,j;
#ifdef USE_BAND_DECOMP
  finteger ld = min(max(roco,12+6*bandwith),lda);
#else
  finteger ld = roco;
#endif
  finteger incld1=ld+1;
  finteger *perm,info;
  Complex *temp,*unused,*a,*d,*e,*expvar,*x;

#define A_org(I,J)   a_org[(I) + (J)*lda]
#define Res(I,J)     res[(I) + (J)*ldres]
#define A(I,J)       a[(I) + (J)*ld]
#define D(I,J)       d[(I) + (J)*ld]
#define E(I,J)       e[(I) + (J)*ld]
#define Exp(I,J)     expvar[(I) + (J)*ld]
#define Unused(I,J)  unused[(I) + (J)*ld]
#define X(I,J)       x[(I) + (J)*ld]

  {
    /* memory management ... */
    /* the max here guarantees that enough workspace */
    /* for the QR decomposition routine is in every */
    /* single matrix */
    int matsize = max(ld,3)*roco;

    ShareWorkspace(a,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(x,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(temp,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(d,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(e,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(unused,sizeof(Complex)*matsize,Work);
    DistributeWorkspace(perm,sizeof(finteger)*ld,Work);
  }

  Exp4Message1((DEBUG_FILE,"Used pade app. for calc. "
	"Exp4PhimHs(Z)=(exp(Z)-1)/Z fac=%e s=%d\n",fac,s));

  /* the scaling */
  fac = ldexp(fac,-s);

  /* copying A_org because A_org is possibly needed  later times */
  placpy ("Full",&roco,&roco,&A_org(0,0),&lda,&A(0,0),&ld);

  /* multiplying A with fac */
  for (j=0;j<roco;j++)
  {
    finteger xstart = max(j - bandwith,0);
    finteger tmpm = min(roco,j+2) - xstart;

    prscal (&tmpm,&fac,&A(xstart,j),&inc1);
  }

  /* Pade approximation for (exp(z)-1)/z */

  /* initialize iteration */

  /* copying A into D,E */
  placpy ("full",&roco,&roco,&A(0,0),&ld,&E(0,0),&ld);
  placpy ("full",&roco,&roco,&A(0,0),&ld,&D(0,0),&ld);

  /* multiplying D,E with pade_d[0],pade_e[0] */
  for (j=0;j<roco;j++)
  {
    finteger xstart = max(j - bandwith,0);
    finteger tmpm = min(roco,j+2) - xstart;

    prscal (&tmpm,&pade_n[0],&E(xstart,j),&inc1);
    prscal (&tmpm,&pade_d[0],&D(xstart,j),&inc1);
  }

  /* adding the identity matrix to E and D */
  paxpy (&roco,&wone,&wone,&inc0,&E(0,0),&incld1);
  paxpy (&roco,&wone,&wone,&inc0,&D(0,0),&incld1);


  /* the Iteration */

  for (i=1;i<PADE_N;i++)
  {
    /* X=X*A */
    bandedmm(roco,&Unused(0,0),ld,&A(0,0),ld,bandwith,1,
	  &X(0,0),ld,min(bandwith*i,roco-1),min(i,roco-1));

    if (i != 1)  /* if it is the first step &X(0,0) points to A */
      temp=x;    /* so new X is in Unused and temp points to an empty array. */
    x=unused;    /* (in the first step temp points to an empty array by */
    unused=temp; /* initialization in the other steps it points to that array
                    that becomes empty */

    /* E=E+pade_n[k]*X */
    /* D=D+pade_d[k]*X */
    for (j=0;j<roco;j++)
    {
      finteger tmpm = min(roco,j+2+i);
      finteger xstart = max(j - (bandwith)*(i+1),0);
      tmpm-=xstart;
      praxpy (&tmpm,&pade_n[i],&X(xstart,j),&inc1,&E(xstart,j),&inc1);
      praxpy (&tmpm,&pade_d[i],&X(xstart,j),&inc1,&D(xstart,j),&inc1);
    }
  } /* for */


  /* last step of Pade approximation: the right division */

#ifdef USE_BAND_DECOMP
  if ((12+6*bandwith+1) < incld1)
  {
    finteger u = min(6*bandwith,roco-1);
    finteger l = min(6,roco-1);
    pgbsv (&roco,&l,&u,&roco,&D(-l-u,0),&incld1,&perm[0],&E(0,0),&ld,&info);
    lapack_err(info,"Exp4PhimHsPade","gesv");
  }
  else
#endif
  {
#ifdef USE_QR_DECOMP_IN_PADE
    Complex *tau=&X(0,0); /* array X is not used in the moment */
    finteger lwork = roco*roco;
    Complex *dwork = &Unused(0,0);
    flogical flag = FFALSE;

    for (i=0;i<roco;perm[i++]=0);
    pgeqpf (&roco,&roco,&D(0,0),&ld,perm,tau,dwork,&info);
    lapack_err(info,"Exp4PhimHsPade","geqpf");

    pormqr ("Left",PORMQR_TRANSPOSE,&roco,&roco,&roco,&D(0,0),&ld,tau,
	  &E(0,0),&ld,dwork,&lwork,&info);
    lapack_err(info,"Exp4PhimHsPade","ormqr");

    ptrtrs ("Upper","No transpose","No Unit",&roco,&roco,&D(0,0),&ld,
	  &E(0,0),&ld,&info);
    lapack_err(info,"Exp4PhimHsPade","trtrs");

    for (i=0;i<roco;i++)
      plapmt (&flag,&inc1,&roco,&E(0,i),&inc1,perm);
#else
    pgesv (&roco,&roco,&D(0,0),&ld,&perm[0],&E(0,0),&ld,&info);
    lapack_err(info,"Exp4PhimHsPade","gesv");
#endif
  } /* the brackets are important !!! */

  if (0<s)  /* Undo Scaling. */
  {
    Complex half = ComplexConst(0.5);

    /* calculates Exp=A*E , initialize the iteration */
    for (j=0;j<roco;j++)
      pgbmv ("N",&roco,&roco,&inc1,&bandwith,&wone,&A(-bandwith,0),&incld1,
	    &E(0,j),&inc1,&wzero,&Unused(0,j),&inc1);

    expvar=unused;
    unused=d;   /* the values in D never used */

    /* adding the identity matrix to Exp */
    paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);


    for (i=1;i<s;i++)  /* The Iteration */
    {
      /* calculates E=0.5*(Exp+Id)*E */

      /* save the Diagonal elements */
      pcopy (&roco,&Exp(0,0),&incld1,&Res(0,0),&inc1);

      /* adds Id to Exp */
      paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);

      /* Unused=0.5*(Exp+id)*E */
      pgemm ("N","N",&roco,&roco,&roco,&half,&Exp(0,0),&ld,
	    &E(0,0),&ld,&wzero,&Unused(0,0),&ld);

      /* rebuild Exp */
      pcopy (&roco,&Res(0,0),&inc1,&Exp(0,0),&incld1);

      temp=e;  /* memory management .... */
      e=unused;
      unused=temp;

      /* calculates Exp=Exp*Exp */
      pgemm ("N","N",&roco,&roco,&roco,&wone,&Exp(0,0),&ld,
	    &Exp(0,0),&ld,&wzero,&Unused(0,0),&ld);

      temp=expvar;  /* memory management .... */
      expvar=unused;
      unused=temp;
    }       /* nearly end of iteration, */


    /*  the last step follows now: */


    /* calculates the return value Res=0.5*(Exp+Id)*E */

    /* adds id to Exp */
    paxpy (&roco,&wone,&wone,&inc0,&Exp(0,0),&incld1);

    /* calculates Res=0.5*Exp*E */
    pgemm ("N","N",&roco,&roco,&roco,&half,&Exp(0,0),&ld,
	  &E(0,0),&ld,&wzero,&Res(0,0),&ldres);
  }
  else /* copying if no scaling was done */
    placpy ("Full",&roco,&roco,&E(0,0),&ld,&Res(0,0),&ldres);

  return 1;
#undef A_org
#undef Res
#undef A
#undef D
#undef E
#undef Exp
#undef Unused
#undef X
}

static void bandedmm(finteger roco,Complex *res,finteger ldres,
      Complex *a,finteger lda,finteger bandau,finteger bandal,
      Complex *b,finteger ldb,finteger bandbu,finteger bandbl)
{
  finteger i;
  finteger ldap1=lda+1;
  finteger height = bandal + bandau;
  finteger astart,bstart;
  finteger tmpm,tmpn;
#define A(I,J) a[(I) + (J)*lda]
#define B(I,J) b[(I) + (J)*ldb]
#define RES(I,J) res[(I) + (J)*ldres]

  for (i=0;i<=min(bandbu,roco);i++)
  {
    tmpn = min(i+bandbl+1,roco);
    tmpm = min(tmpn+bandal,roco);
    pgbmv ("N",&tmpm,&tmpn,&bandal,&bandau,&wone,&A(-bandau,0),&ldap1,
	  &B(0,i),&inc1,&wzero,&RES(0,i),&inc1);
  }

  for (;i<min(bandau+bandbu,roco);i++)
  {
    finteger bstart = i - bandbu;
    finteger l = bandal + bstart;
    finteger u = bandau - bstart;
    tmpn = min(i+bandbl+1,roco) - bstart;
    tmpm = min(tmpn+bandal+bstart,roco);
    pgbmv ("N",&tmpm,&tmpn,&l,&u,&wone,&A(-u,bstart),&ldap1,
	  &B(bstart,i),&inc1,&wzero,&RES(0,i),&inc1);
  }

  for (;i<roco-bandal-bandbl;i++)
  {
    bstart = i - bandbu;
    astart = bstart - bandau;
    tmpn = bandbl + 1 + bandbu;
    tmpm = height + tmpn;
    pgbmv ("N",&tmpm,&tmpn,&height,&inc0,&wone,&A(astart,bstart),&ldap1,
	  &B(bstart,i),&inc1,&wzero,&RES(astart,i),&inc1);
  }

  for (;i<roco;i++)
  {
    bstart = i - bandbu;
    astart = bstart - bandau;
    tmpn = roco - bstart;
    tmpm = roco - astart;
    pgbmv ("N",&tmpm,&tmpn,&height,&inc0,&wone,&A(astart,bstart),&ldap1,
	  &B(bstart,i),&inc1,&wzero,&RES(astart,i),&inc1);
  }
#undef A
#undef B
#undef RES
}
