/*************************************************************************/
/* Copyright (C) 1995-1997 Marlis Hochbruck, Christian Lubich, Hubert    */
/* Selhofer. All rights reserved.                                        */
/*                                                                       */
/* C code written by Mathias Froehlich.                                  */
/*                                                                       */
/* This code is part of a copyrighted package. For details, see the file */
/* "COPYING" in the top-level directory.                                 */
/*************************************************************************/

#include <stdlib.h>
#include <stdarg.h>
#include <math.h>
#include "ftypes.h"
#include "fblas.h"
#include "exp4.h"
#include "prec.h"
#include "errors.h"
#include "debug.h"

#include "work.h"
#include "scnorm.h"

#ifndef DEFAULT_ATOL
#define DEFAULT_ATOL 1e-6
#endif
#ifndef DEFAULT_RTOL
#define DEFAULT_RTOL 1e-3
#endif

/* structure for a tolerance value */
typedef struct {
  Exp4TolType Type;
  Real Value;
  Real *Ptr;
} Tolerance;

/* the two tolerances */
static Tolerance Atol = {Scalar, DEFAULT_ATOL, NULL};
static Tolerance Rtol = {Scalar, DEFAULT_RTOL, NULL};

/* data for the scaled norm functions */
static finteger neq;
static Real *scale;
static Exp4NormType Norm = Norm2;

int PREC(Exp4SetAtol) (Exp4TolType Type, ...)
{
  va_list list;

  va_start(list,Type);

  Atol.Type = Type;
  switch (Type) {
  case Scalar:
    Atol.Value = va_arg(list,Real);
    break;
  case Vector:
    Atol.Ptr = va_arg(list,Real *);
    break;
  case Default:
    Atol.Type = Scalar;
    Atol.Value = DEFAULT_ATOL;
    break;
  default:
    Warning("Exp4SetAtol Warning: Illegal Type Value!");
    va_end(list);
    return 1;
  }

  va_end(list);
  return 0;
}

int PREC(Exp4SetRtol) (Exp4TolType Type, ...)
{
  va_list list;

  va_start(list,Type);

  Rtol.Type = Type;
  switch (Type) {
  case Scalar:
    Rtol.Value = va_arg(list,Real);
    break;
  case Vector:
    Rtol.Ptr = va_arg(list,Real *);
    break;
  case Default:
    Rtol.Type = Scalar;
    Rtol.Value = DEFAULT_RTOL;
    break;
  default:
    Warning("Exp4SetRtol Warning: Illegal Type Value!");
    va_end(list);
    return 1;
  }

  va_end(list);
  return 0;
}

void PREC(Exp4SetNormType) (Exp4NormType NormIn)
{
  Norm = NormIn;
}

void PREC(Exp4InitScNorm) (Complex *First,finteger neqIn)
     /* initializes variables for using scaled norms */
{
  neq = neqIn;
  scale = Exp4GetWorkPointer(neq*sizeof(Real));

  if (Atol.Type == Scalar)
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr;
      ScPtr = scale;

      for (i=0;i<neq;i++) {
	Complex Tmp = *First++;
	*ScPtr++ = 1.0/(Atol.Value + Rtol.Value * pabs(Tmp));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*RtolPtr;
      ScPtr = scale; RtolPtr = Rtol.Ptr;

      for (i=0;i<neq;i++) {
	Complex Tmp = *First++;
	*ScPtr++ = 1.0/(Atol.Value + (*RtolPtr++) * pabs(Tmp));
      }
    }
  }
  else
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr,*AtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Tmp = *First++;
	*ScPtr++ = 1.0/((*AtolPtr++) + Rtol.Value * pabs(Tmp));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*AtolPtr,*RtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr; RtolPtr = Rtol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Tmp = *First++;
	*ScPtr++ = 1.0/((*AtolPtr++) + (*RtolPtr++) * pabs(Tmp));
      }
    }
  }
}

void PREC(Exp4SetScale) (Complex *x1,Complex *x2)
     /* Generates a new vektor scale by the max(abs(x1),abs(x2)) */
{
  if (Atol.Type == Scalar)
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr;
      ScPtr = scale;

      for (i=0;i<neq;i++) {
	Complex Val1 = *x1++,Val2 = *x2++;
	Real Abs1 = pabs(Val1),Abs2 = pabs(Val2);
	*ScPtr++ = 1.0/(Atol.Value + Rtol.Value * max(Abs1,Abs2));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*RtolPtr;
      ScPtr = scale; RtolPtr = Rtol.Ptr;

      for (i=0;i<neq;i++) {
	Complex Val1 = *x1++,Val2 = *x2++;
	Real Abs1 = pabs(Val1),Abs2 = pabs(Val2);
	*ScPtr++ = 1.0/(Atol.Value + (*RtolPtr++) * max(Abs1,Abs2));
      }
    }
  }
  else
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr,*AtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Val1 = *x1++,Val2 = *x2++;
	Real Abs1 = pabs(Val1),Abs2 = pabs(Val2);
	*ScPtr++ = 1.0/((*AtolPtr++) + Rtol.Value * max(Abs1,Abs2));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*AtolPtr,*RtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr; RtolPtr = Rtol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Val1 = *x1++,Val2 = *x2++;
	Real Abs1 = pabs(Val1),Abs2 = pabs(Val2);
	*ScPtr++ = 1.0/((*AtolPtr++) + (*RtolPtr++) * max(Abs1,Abs2));
      }
    }
  }
}

void PREC(Exp4SetScale1) (Complex *x1)
     /* Generates a new vektor scale by the abs(x1) */
{
  if (Atol.Type == Scalar)
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr;
      ScPtr = scale;

      for (i=0;i<neq;i++) {
	Complex Tmp = *x1++;
	*ScPtr++ = 1.0/(Atol.Value + Rtol.Value * pabs(Tmp));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*RtolPtr;
      ScPtr = scale; RtolPtr = Rtol.Ptr;

      for (i=0;i<neq;i++) {
	Complex Tmp = *x1++;
	*ScPtr++ = 1.0/(Atol.Value + (*RtolPtr++) * pabs(Tmp));
      }
    }
  }
  else
  {
    if (Rtol.Type == Scalar)
    {
      int i;
      Real *ScPtr,*AtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Tmp = *x1++;
	*ScPtr++ = 1.0/((*AtolPtr++) + Rtol.Value * pabs(Tmp));
      }
    }
    else
    {
      int i;
      Real *ScPtr,*AtolPtr,*RtolPtr;
      ScPtr = scale; AtolPtr = Atol.Ptr; RtolPtr = Rtol.Ptr;
      
      for (i=0;i<neq;i++) {
	Complex Tmp = *x1++;
	*ScPtr++ = 1.0/((*AtolPtr++) + (*RtolPtr++) * pabs(Tmp));
      }
    }
  }
}

Real PREC(Exp4ScaledNorm) (Complex *x)
     /* Returns the scaled Norm of x. */
{  
  Real *ScPtr;
  ScPtr = scale;

  if (Norm == Norm2)
  {
    int i;
    Real Sum = 0.0;

    for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
      Real Tmp = (*x) * (*ScPtr);
      Sum += Tmp*Tmp;
#else
      Real TmpR = (*x).r * (*ScPtr);
      Real TmpI = (*x).i * (*ScPtr);
      Sum += TmpR*TmpR+TmpI*TmpI;
#endif
      x++; ScPtr++;
    }

    return sqrt(Sum/neq);
  }
  else
  {
    int i;
    Real Max = 0.0;

    for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
      Real Tmp = (*x) * (*ScPtr);
      if (Max < Tmp) Max = Tmp;
#else
      Real TmpR = (*x).r * (*ScPtr);
      Real TmpI = (*x).i * (*ScPtr);
      Real Tmp = TmpR*TmpR+TmpI*TmpI;
      if (Max < Tmp) Max = Tmp;
#endif
      x++; ScPtr++;
    }

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    return Max;
#else
    /* note: here the sqares of the value are compared to
     * avoid expensive sqrt() */
    return sqrt(Max);
#endif
  }
}

Real PREC(Exp4ScaledNormTmpScale) (Complex *x,
      Complex *Tmpscale)
     /* returns the scaled norm of x taking the scale abs(Tmpscale) */
{
  if (Norm == Norm2)
  {
    Real Sum = 0.0;

    if (Atol.Type == Scalar)
    {
      if (Rtol.Type == Scalar)
      {
	int i;
	Real *ScPtr;
	ScPtr = scale;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x) / (Atol.Value + Rtol.Value*pabs(*Tmpscale));
	  Sum += Tmp*Tmp;
#else
	  Real Sc = 1.0 / (Atol.Value + Rtol.Value*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Sum += TmpR*TmpR+TmpI*TmpI;
#endif
	  x++; Tmpscale++;
	}
      }
      else
      {
	int i;
	Real *ScPtr,*RtolPtr;
	ScPtr = scale; RtolPtr = Rtol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x) / (Atol.Value+(*RtolPtr++)*pabs(*Tmpscale));
	  Sum += Tmp*Tmp;
#else
	  Real Sc = 1.0 / (Atol.Value+(*RtolPtr++)*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Sum += TmpR*TmpR+TmpI*TmpI;
#endif
	  x++; Tmpscale++;
	}
      }
    }
    else
    {
      if (Rtol.Type == Scalar)
      {
	int i;
	Real *ScPtr,*AtolPtr;
	ScPtr = scale; AtolPtr = Atol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x)/((*AtolPtr++)+Rtol.Value*pabs(*Tmpscale));
	  Sum += Tmp*Tmp;
#else
	  Real Sc = 1.0/((*AtolPtr++)+Rtol.Value*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Sum += TmpR*TmpR+TmpI*TmpI;
#endif
	  x++; Tmpscale++;
	}
      }
      else
      {
	int i;
	Real *ScPtr,*AtolPtr,*RtolPtr;
	ScPtr = scale; AtolPtr = Atol.Ptr; RtolPtr = Rtol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x)/((*AtolPtr++)+(*RtolPtr++)*pabs(*Tmpscale));
	  Sum += Tmp*Tmp;
#else
	  Real Sc = 1.0/((*AtolPtr++)+(*RtolPtr++)*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Sum += TmpR*TmpR+TmpI*TmpI;
#endif
	  x++; Tmpscale++;
	}
      }
    }

    return sqrt(Sum/neq);
  }
  else
  {
    Real Max = 0.0;

    if (Atol.Type == Scalar)
    {
      if (Rtol.Type == Scalar)
      {
	int i;
	Real *ScPtr;
	ScPtr = scale;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x) / (Atol.Value + Rtol.Value*pabs(*Tmpscale));
	  if (Max < Tmp) Max = Tmp;
#else
	  Real Sc = 1.0 / (Atol.Value + Rtol.Value*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Real Tmp = TmpR*TmpR+TmpI*TmpI;
	  if (Max < Tmp) Max = Tmp;
#endif
	  x++; Tmpscale++;
	}
      }
      else
      {
	int i;
	Real *ScPtr,*RtolPtr;
	ScPtr = scale; RtolPtr = Rtol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x) / (Atol.Value+(*RtolPtr++)*pabs(*Tmpscale));
	  if (Max < Tmp) Max = Tmp;
#else
	  Real Sc = 1.0 / (Atol.Value+(*RtolPtr++)*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Real Tmp = TmpR*TmpR+TmpI*TmpI;
	  if (Max < Tmp) Max = Tmp;
#endif
	  x++; Tmpscale++;
	}
      }
    }
    else
    {
      if (Rtol.Type == Scalar)
      {
	int i;
	Real *ScPtr,*AtolPtr;
	ScPtr = scale; AtolPtr = Atol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x) / ((*AtolPtr++)+Rtol.Value*pabs(*Tmpscale));
	  if (Max < Tmp) Max = Tmp;
#else
	  Real Sc = 1.0 / ((*AtolPtr++)+Rtol.Value*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Real Tmp = TmpR*TmpR+TmpI*TmpI;
	  if (Max < Tmp) Max = Tmp;
#endif
	  x++; Tmpscale++;
	}
      }
      else
      {
	int i;
	Real *ScPtr,*AtolPtr,*RtolPtr;
	ScPtr = scale; AtolPtr = Atol.Ptr; RtolPtr = Rtol.Ptr;
	
	for (i=0;i<neq;i++) {
#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
	  Real Tmp = (*x)/((*AtolPtr++)+(*RtolPtr++)*pabs(*Tmpscale));
	  if (Max < Tmp) Max = Tmp;
#else
	  Real Sc = 1.0/((*AtolPtr++)+(*RtolPtr++)*pabs(*Tmpscale));
	  Real TmpR = (*x).r * Sc;
	  Real TmpI = (*x).i * Sc;
	  Real Tmp = TmpR*TmpR+TmpI*TmpI;
	  if (Max < Tmp) Max = Tmp;
#endif
	  x++; Tmpscale++;
	}
      }
    }

#if (defined(SINGLE_PRECISION) || defined(DOUBLE_PRECISION))
    return Max;
#else
    /* note: here the sqares of the value are compared to
     * avoid expensive sqrt() */
    return sqrt(Max);
#endif
  }
}

Real PREC(Exp4GetMinRtol) ()
     /* returns the minimal rtol value, mainly for statistics */
{
  if (Rtol.Type == Scalar)
    return Rtol.Value;
  else
  {
    finteger i;
    Real Min = *Rtol.Ptr;

    for (i=1;i<neq;i++)
      if (Rtol.Ptr[i] < Min)
	Min = Rtol.Ptr[i];
    return Min;
  }
}

Real PREC(Exp4GetMaxRtol) ()
     /* returns the maximal rtol value, mainly for statistics */
{
  if (Rtol.Type == Scalar)
    return Rtol.Value;
  else
  {
    finteger i;
    Real Max = *Rtol.Ptr;

    for (i=1;i<neq;i++)
      if (Rtol.Ptr[i] > Max)
	Max = Rtol.Ptr[i];
    return Max;
  }
}

Real PREC(Exp4GetMinAtol) ()
     /* returns the minimal atol value, mainly for statistics */
{
  if (Atol.Type == Scalar)
    return Atol.Value;
  else
  {
    finteger i;
    Real Min = *Atol.Ptr;

    for (i=1;i<neq;i++)
      if (Atol.Ptr[i] < Min)
	Min = Atol.Ptr[i];
    return Min;
  }
}

Real PREC(Exp4GetMaxAtol) ()
     /* returns the maximal atol value, mainly for statistics */
{
  if (Atol.Type == Scalar)
    return Atol.Value;
  else
  {
    finteger i;
    Real Max = *Atol.Ptr;

    for (i=1;i<neq;i++)
      if (Atol.Ptr[i] > Max)
	Max = Atol.Ptr[i];
    return Max;
  }
}
